/******************************************************************************/
/*                                                                            */
/*  <<< Copyright >>>                                                         */
/*                                                                            */
/*  The "xmemo" software is copyright 1999 by Takao Abe.                      */
/*  All rights reserved.                                                      */
/*                                                                            */
/*  The "xmemo" software refers to the source code, the binary code,          */
/*  the documents, and all derivatives based on this software under           */
/*  the copyright law, that is a software containing a program or a           */
/*  portion of a program, regardless of modifications and/or translations.    */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  <<< Licence and redistribution >>>                                        */
/*                                                                            */
/*  The source code and the binary code are free software.                    */
/*  The use and redistribution of source code and binary code are permitted   */
/*  provided that the following conditions are met:                           */
/*  You may use the software for private use and commercial use,              */
/*  and may modify your copy or copies of the software or any portion of it.  */
/*                                                                            */
/*  THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESSED OR IMPLIED            */
/*  WARRANTIES OF ANY KIND, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED        */
/*  WARRANTIES OF MERCHANTABLILITY AND FITNESS FOR A PARTICULAR PURPOSE.      */
/*  IN NO EVENT SHALL THE AUTHOR TAKAO ABE BE LIABLE FOR ANY DIRECT,          */
/*  INDIRECT, GENERAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES           */
/*  ( INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR       */
/*  SERVICES; LOSS OF USE, DATA OR PROFITS; OR BUSINESS INTERRUPTION )        */
/*  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,       */
/*  STRICT LIABILITY, OR TORT ( INCLUDING NEGLIGENCE OR OTHERWISE )           */
/*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED       */
/*  OF THE POSSIBILITY OF SUCH DAMAGE.                                        */
/*                                                                            */
/*  Redistribution and use in source form or binary form, with or without     */
/*  modification, the whole program or a portion of it, are permitted         */
/*  provided that one retains the entire copyright notice properly, and       */
/*  displays or retains this license in the documentation and/or other        */
/*  materials provided with the distribution.                                 */
/*                                                                            */
/*  Redistribution and use in source form or binary form, with modification,  */
/*  the whole program or a portion of it, are permitted provided that one     */
/*  declares that it is modified software and not original, and displays or   */
/*  describes any modifications in the documentation and/or other materials   */
/*  provided with the distribution.                                           */
/*                                                                            */
/*  This software and the name of the author must not be used to endorse or   */
/*  promote products derived from this software without prior written         */
/*  permission.                                                               */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  EMail   abetakao@bea.hi-ho.ne.jp                                          */
/*          HCA00763@nifty.ne.jp                                              */
/*                                                                            */
/*  URL     http://www.bea.hi-ho.ne.jp/abetakao/                              */
/*                                                                            */
/******************************************************************************/


#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<time.h>


#include	<X11/Intrinsic.h>
#include	<X11/StringDefs.h>
#include	<X11/Shell.h>

#include	<X11/Xaw/AsciiText.h>
#include	<X11/Xaw/Box.h>
#include	<X11/Xaw/Command.h>
#include	<X11/Xaw/Form.h>
#include	<X11/Xaw/Label.h>
#include	<X11/Xaw/List.h>
#include	<X11/Xaw/Toggle.h>
#include	<X11/Xaw/Viewport.h>


#include	"xmemo.h"
#include	"xmemodata.h"
#include	"xmemofunc.h"


/******************************************************************************/
/*  Local variables                                                           */
/******************************************************************************/


static	Widget	wDlgOptionNetwork ;
static	Widget	wDlgOptionNetworkForm ;
static	Widget	wDlgOptionNetworkEdit ;
static	Widget	wDlgOptionNetworkViewport ;
static	Widget	wDlgOptionNetworkList ;
static	Widget	wDlgOptionNetworkLblProtocol ;
static	Widget	wDlgOptionNetworkTglProtocol [ 3 ] ;
static	Widget	wDlgOptionNetworkLblPort ;
static	Widget	wDlgOptionNetworkTxtPort ;
static	Widget	wDlgOptionNetworkLblAddress [ 2 ] ;
static	Widget	wDlgOptionNetworkTglAddress [ 3 ] ;
static	Widget	wDlgOptionNetworkTxtAddress [ 2 ] ;
static	Widget	wDlgOptionNetworkCmdDelete ;
static	Widget	wDlgOptionNetworkCmdEnable ;
static	Widget	wDlgOptionNetworkCmdDisable ;
static	Widget	wDlgOptionNetworkCmdClose ;
static	Widget	wDlgOptionNetworkCmdCancel ;


static	int 	iListCount ;
static	int 	iListSelect ;
static	char	**pListArray ;
static	zNetworkType	**pNetworkArray ;


/******************************************************************************/
/*  xmemo_FreeNetworkEntry                                                    */
/******************************************************************************/

static void xmemo_FreeNetworkEntry ( zNetworkType *pNetwork )
{

	if ( pNetwork->pAddr[0] != NULL ) XtFree ( (void*) ( pNetwork->pAddr[0] ) ) ;
	if ( pNetwork->pAddr[1] != NULL ) XtFree ( (void*) ( pNetwork->pAddr[1] ) ) ;
	if ( pNetwork->pPort    != NULL ) XtFree ( (void*) ( pNetwork->pPort    ) ) ;
	if ( pNetwork->pLine    != NULL ) XtFree ( (void*) ( pNetwork->pLine    ) ) ;
	XtFree ( (void*) pNetwork ) ;

}

/******************************************************************************/
/*  xmemo_CopyNetworkEntry                                                    */
/******************************************************************************/

static zNetworkType* xmemo_CopyNetworkEntry ( zNetworkType *pNetwork )
{

zNetworkType	*pCopy ;

	pCopy = (zNetworkType*) XtCalloc ( sizeof(zNetworkType), 1 ) ;
	if ( pCopy == NULL ) {
		xmemo_ErrorLog ( "xmemo_CopyNetworkEntry", "XtCalloc", NULL ) ;
		return NULL ;
	}

	if ( pNetwork->pPort != NULL ) {
		pCopy->pPort = XtMalloc ( strlen ( pNetwork->pPort ) + 1 ) ;
		if ( pCopy->pPort == NULL ) {
			xmemo_ErrorLog ( "OptionNetworkAdd_Click", "XtMalloc", NULL ) ;
			xmemo_FreeNetworkEntry ( pCopy ) ;
			return NULL ;
		}
		strcpy ( pCopy->pPort, pNetwork->pPort ) ;
	}

	if ( pNetwork->pAddr[0] != NULL ) {
		pCopy->pAddr[0] = XtMalloc ( strlen ( pNetwork->pAddr[0] ) + 1 ) ;
		if ( pCopy->pAddr[0] == NULL ) {
			xmemo_ErrorLog ( "OptionNetworkAdd_Click", "XtMalloc", NULL ) ;
			xmemo_FreeNetworkEntry ( pCopy ) ;
			return NULL ;
		}
		strcpy ( pCopy->pAddr[0], pNetwork->pAddr[0] ) ;
	}

	if ( pNetwork->pAddr[1] != NULL ) {
		pCopy->pAddr[1] = XtMalloc ( strlen ( pNetwork->pAddr[1] ) + 1 ) ;
		if ( pCopy->pAddr[1] == NULL ) {
			xmemo_ErrorLog ( "OptionNetworkAdd_Click", "XtMalloc", NULL ) ;
			xmemo_FreeNetworkEntry ( pCopy ) ;
			return NULL ;
		}
		strcpy ( pCopy->pAddr[1], pNetwork->pAddr[1] ) ;
	}

	pCopy->cProtocol = pNetwork->cProtocol ;
	pCopy->cSearch   = pNetwork->cSearch  ;
	pCopy->bEnable   = pNetwork->bEnable  ;
	pCopy->iAddr[0]  = pNetwork->iAddr[0] ;
	pCopy->iAddr[1]  = pNetwork->iAddr[1] ;
	pCopy->iPort     = pNetwork->iPort    ;
	pCopy->bActive   = pNetwork->bActive  ;
	pCopy->pUdpSock  = pNetwork->pUdpSock ;

	return pCopy ;

}

/******************************************************************************/
/*  xmemo_SetNetworkListLine                                                  */
/******************************************************************************/

static void xmemo_SetNetworkListLine ( zNetworkType *pNetwork )
{

char	sBuf [ 36 ] ;

	pNetwork->pLine = XtMalloc ( 66 ) ;
	if ( pNetwork->pLine == NULL ) {
		xmemo_ErrorLog ( "xmemo_SetNetworkListLine", "XtMalloc", NULL ) ;
		return ;
	}

	/* Length = 9 ( Protocol ) */
	switch ( pNetwork->cProtocol ) {
	case XMEMO_PROTOCOL_IPMSG    : strcpy ( pNetwork->pLine, "IPMsg    " ) ; break ;
	case XMEMO_PROTOCOL_OBOEGAKI : strcpy ( pNetwork->pLine, "Oboegaki " ) ; break ;
	case XMEMO_PROTOCOL_HUSEN    : strcpy ( pNetwork->pLine, "Husenshi " ) ; break ;
	default                      : strcpy ( pNetwork->pLine, "Invalid  " ) ; break ;
	}

	/* Length = 2 ( Spaces ) */
	strcat ( pNetwork->pLine, "  " ) ;

	/* Length = 5 */
	if ( pNetwork->pPort == NULL ) {
		strncpy ( sBuf, "     "       , 5 ) ;
	} else if ( strlen ( pNetwork->pPort ) > 5 ) {
		strncpy ( sBuf, pNetwork->pPort, 5 ) ;
	} else {
		sprintf ( sBuf, "%5s", pNetwork->pPort ) ;
	}
	sBuf[5] = 0 ;
	strcat ( pNetwork->pLine, sBuf ) ;

	/* Length = 2 ( Spaces ) */
	strcat ( pNetwork->pLine, "  " ) ;

	/* Length = 35 ( Address ) */
	if ( pNetwork->cSearch == XMEMO_SEARCH_RANGE ) {
		if ( pNetwork->pAddr[0] == NULL ) {
			sprintf ( sBuf, "%15s", " " ) ;
		} else if ( strlen ( pNetwork->pAddr[0] ) > 15 ) {
			strncpy ( sBuf, pNetwork->pAddr[0], 15 ) ;
		} else {
			sprintf ( sBuf, "%-15s", pNetwork->pAddr[0] ) ;
		}
		sBuf[15] = 0 ;
		strcat ( pNetwork->pLine, sBuf ) ;
		strcat ( pNetwork->pLine, "  -  " ) ;
		if ( pNetwork->pAddr[1] == NULL ) {
			sprintf ( sBuf, "%15s", " " ) ;
		} else if ( strlen ( pNetwork->pAddr[1] ) > 15 ) {
			strncpy ( sBuf, pNetwork->pAddr[1], 15 ) ;
		} else {
			sprintf ( sBuf, "%-15s", pNetwork->pAddr[1] ) ;
		}
		sBuf[15] = 0 ;
		strcat ( pNetwork->pLine, sBuf ) ;
	} else {
		if ( pNetwork->pAddr[0] == NULL ) {
			sprintf ( sBuf, "%35s", " " ) ;
		} else if ( strlen ( pNetwork->pAddr[0] ) > 35 ) {
			strncpy ( sBuf, pNetwork->pAddr[0], 35 ) ;
		} else {
			sprintf ( sBuf, "%-35s", pNetwork->pAddr[0] ) ;
		}
		sBuf[35] = 0 ;
		strcat ( pNetwork->pLine, sBuf ) ;
	}

	/* Length = 2 ( Spaces ) */
	strcat ( pNetwork->pLine, "  " ) ;

	/* Length = 7 ( Enable,Disable ) */
	if ( pNetwork->bEnable ) strcat ( pNetwork->pLine, "Enable " ) ;
	else                    strcat ( pNetwork->pLine, "Disable" ) ;

}

/******************************************************************************/
/*  xmemo_SetNetworkListArray                                                 */
/******************************************************************************/

static void xmemo_SetNetworkListArray ( )
{

zChainType	*pChain ;
zNetworkType	*pNetwork ;
int 	n ;
Arg 	xArg [ 3 ] ;
int 	i, j ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_SetNetworkListArray", NULL ) ;

	if ( pListArray != NULL ) {
		XtFree ( (void*) pListArray ) ;
	}

	if ( pNetworkArray != NULL ) {
		XtFree ( (void*) pNetworkArray ) ;
	}

	/* Count effective entry */

	iListCount = 0 ;
	for ( pChain = zNetworkAnchor.pHead ; pChain != NULL ; pChain = pChain->pNext ) {

		pNetwork = (zNetworkType*) ( pChain->pData ) ;

		if ( ! pNetwork->bDelete ) iListCount ++ ;

	}

	/* Allocate memory */

	pListArray = (char**) XtCalloc ( sizeof(char*), iListCount + 1 ) ;
	if ( pListArray == NULL ) {
		xmemo_ErrorLog ( "xmemo_SetNetworkListArray", "XtCalloc", NULL ) ;
		return ;
	}

	pNetworkArray = (zNetworkType**) XtCalloc ( sizeof(zNetworkType*), iListCount + 1 ) ;
	if ( pNetworkArray == NULL ) {
		xmemo_ErrorLog ( "xmemo_SetNetworkListArray", "XtCalloc", NULL ) ;
		return ;
	}

	/* Set list entry ( Sorted by IP address ) */

	iListCount = 0 ;
	for ( pChain = zNetworkAnchor.pHead ; pChain != NULL ; pChain = pChain->pNext ) {

		pNetwork = (zNetworkType*) ( pChain->pData ) ;

		if ( ! pNetwork->bDelete ) {

			if ( pNetwork->pLine == NULL ) xmemo_SetNetworkListLine ( pNetwork ) ;
			if ( pNetwork->pLine == NULL ) continue ;
			for ( i = 0 ; i < iListCount ; i ++ ) {
				if ( pNetworkArray[i]->iAddr[0] >  pNetwork->iAddr[0]
				  || pNetworkArray[i]->iAddr[0] == pNetwork->iAddr[0]
				  && pNetworkArray[i]->iPort    >  pNetwork->iPort    ) break ;
			}
			for ( j = iListCount - 1 ; j >= i ; j -- ) {
				pListArray[j+1] = pListArray[j] ;
				pNetworkArray[j+1] = pNetworkArray[j] ;
			}
			pListArray[i] = pNetwork->pLine ;
			pNetworkArray[i] = pNetwork ;
			iListCount ++ ;

		}

	}
	pListArray[iListCount] = NULL ;

	/* Refresh list widget */

	n = 0 ;
	XtSetArg ( xArg[n], XtNlist, pListArray ) ; n ++ ;
	XtSetArg ( xArg[n], XtNnumberStrings, iListCount ) ; n ++ ;
	XtSetValues ( wDlgOptionNetworkList, xArg, n ) ;
	XawListChange ( wDlgOptionNetworkList, pListArray, iListCount, 0, False ) ;
	iListSelect = -1 ;

}

/******************************************************************************/
/*  OptionNetworkAdd_Click                                                    */
/******************************************************************************/

static void OptionNetworkAdd_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

int 	i, n ;
Arg 	xArg [ 10 ] ;
int 	bState ;
int 	bTglProtocol [ 3 ] ;
char	*pTxtPort ;
int 	bTglAddress [ 3 ] ;
char	*pTxtAddress [ 2 ] ;

char	cTglProtocol ;
char	cTglAddress ;

zNetworkType	*pNetwork ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "OptionNetworkAdd_Click", NULL ) ;

	/* Input values */

	for ( i = 0 ; i < 3 ; i ++ ) {
		n = 0 ;
		XtSetArg ( xArg[n], XtNstate, &bState ) ; n ++ ;
		XtGetValues ( wDlgOptionNetworkTglProtocol[i], xArg, n ) ;
		bTglProtocol[i] = ( ( bState & True ) == True ) ;
	}
	if      ( bTglProtocol[0] ) cTglProtocol = XMEMO_PROTOCOL_IPMSG ;
	else if ( bTglProtocol[1] ) cTglProtocol = XMEMO_PROTOCOL_OBOEGAKI ;
	else if ( bTglProtocol[2] ) cTglProtocol = XMEMO_PROTOCOL_HUSEN ;
	else                        cTglProtocol = XMEMO_PROTOCOL_INVALID ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNstring, &pTxtPort ) ; n ++ ;
	XtGetValues ( wDlgOptionNetworkTxtPort, xArg, n ) ;

	for ( i = 0 ; i < 3 ; i ++ ) {
		n = 0 ;
		XtSetArg ( xArg[n], XtNstate, &bState ) ; n ++ ;
		XtGetValues ( wDlgOptionNetworkTglAddress[i], xArg, n ) ;
		bTglAddress[i] = ( ( bState & True ) == True ) ;
	}
	if      ( bTglAddress[0] ) cTglAddress = XMEMO_SEARCH_BROADCAST ;
	else if ( bTglAddress[1] ) cTglAddress = XMEMO_SEARCH_HOST ;
	else if ( bTglAddress[2] ) cTglAddress = XMEMO_SEARCH_RANGE ;
	else                       cTglAddress = XMEMO_SEARCH_INVALID ;

	for ( i = 0 ; i < 2 ; i ++ ) {
		n = 0 ;
		XtSetArg ( xArg[n], XtNstring, &(pTxtAddress[i]) ) ; n ++ ;
		XtGetValues ( wDlgOptionNetworkTxtAddress[i], xArg, n ) ;
	}

	/* Check */

	if ( cTglProtocol == XMEMO_PROTOCOL_INVALID
	  || cTglAddress  == XMEMO_SEARCH_INVALID ) {
		XBell ( XtDisplay(wTop), 100 ) ;
		return ;
	}

	/* Allocate memory */

	pNetwork = (zNetworkType*) XtCalloc ( sizeof(zNetworkType), 1 ) ;
	if ( pNetwork == NULL ) {
		xmemo_ErrorLog ( "OptionNetworkAdd_Click", "XtCalloc", NULL ) ;
		return ;
	}

	pNetwork->cProtocol = cTglProtocol ;

	pNetwork->cSearch   = cTglAddress  ;

	if ( pTxtPort != NULL ) {
		pNetwork->pPort = XtMalloc ( strlen ( pTxtPort ) + 1 ) ;
		if ( pNetwork->pPort == NULL ) {
			xmemo_ErrorLog ( "OptionNetworkAdd_Click", "XtMalloc", NULL ) ;
			xmemo_FreeNetworkEntry ( pNetwork ) ;
			return ;
		}
		strcpy ( pNetwork->pPort, pTxtPort ) ;
		pNetwork->iPort = atoi ( pNetwork->pPort ) ;
	}

	if ( pTxtAddress[0] != NULL ) {
		pNetwork->pAddr[0] = XtMalloc ( strlen ( pTxtAddress[0] ) + 1 ) ;
		if ( pNetwork->pAddr[0] == NULL ) {
			xmemo_ErrorLog ( "OptionNetworkAdd_Click", "XtMalloc", NULL ) ;
			xmemo_FreeNetworkEntry ( pNetwork ) ;
			return ;
		}
		strcpy ( pNetwork->pAddr[0], pTxtAddress[0] ) ;
		pNetwork->iAddr[0] = ipaddress_string_to_long ( pNetwork->pAddr[0] ) ;
	}

	if ( pTxtAddress[1] != NULL  &&  cTglAddress == XMEMO_SEARCH_RANGE ) {
		pNetwork->pAddr[1] = XtMalloc ( strlen ( pTxtAddress[1] ) + 1 ) ;
		if ( pNetwork->pAddr[1] == NULL ) {
			xmemo_ErrorLog ( "OptionNetworkAdd_Click", "XtMalloc", NULL ) ;
			xmemo_FreeNetworkEntry ( pNetwork ) ;
			return ;
		}
		strcpy ( pNetwork->pAddr[1], pTxtAddress[1] ) ;
		pNetwork->iAddr[1] = ipaddress_string_to_long ( pNetwork->pAddr[1] ) ;
	}

	pNetwork->bEnable = TRUE ;
	pNetwork->bActive = FALSE ;
	pNetwork->bDelete = FALSE ;
	pNetwork->bInsert = TRUE ;

	/* Insert into chained list */

	pNetwork->pChain = xmemo_InsertIntoChain ( &zNetworkAnchor, pNetwork ) ;
	if ( pNetwork->pChain == NULL ) {
		xmemo_ErrorLog ( "xmemo_CreateMemo", "xmemo_InsertIntoChain", NULL ) ;
		xmemo_FreeNetworkEntry ( pNetwork ) ;
		return ;
	}

	/* Refresh list */

	xmemo_SetNetworkListArray ( ) ;

}

static XtCallbackRec pOptionNetworkAdd_CallbackRec [ ] =
{
	{ (XtCallbackProc) OptionNetworkAdd_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  OptionNetworkDelete_Click                                                 */
/******************************************************************************/

static void OptionNetworkDelete_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

zNetworkType		*pNetwork ;
int 	n ;
Arg 	xArg [ 16 ] ;

	if ( iListSelect < 0 || iListCount <= iListSelect ) return ;
	if ( bMemoTraceLog ) xmemo_TraceLog ( "OptionNetworkDelete_Click", NULL ) ;

	pNetwork = pNetworkArray[iListSelect] ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNstate, True ) ; n ++ ;
	switch ( pNetwork->cProtocol ) {
	case XMEMO_PROTOCOL_IPMSG :
		XtSetValues ( wDlgOptionNetworkTglProtocol[0], xArg, n ) ;
		break ;
	case XMEMO_PROTOCOL_OBOEGAKI :
		XtSetValues ( wDlgOptionNetworkTglProtocol[1], xArg, n ) ;
		break ;
	case XMEMO_PROTOCOL_HUSEN :
		XtSetValues ( wDlgOptionNetworkTglProtocol[2], xArg, n ) ;
		break ;
	}

	n = 0 ;
	XtSetArg ( xArg[n], XtNstring, pNetwork->pPort ) ; n ++ ;
	XtSetValues ( wDlgOptionNetworkTxtPort, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNstate, True ) ; n ++ ;
	switch ( pNetwork->cSearch ) {
	case XMEMO_SEARCH_BROADCAST :
		XtSetValues ( wDlgOptionNetworkTglAddress[0], xArg, n ) ;
		break ;
	case XMEMO_SEARCH_HOST :
		XtSetValues ( wDlgOptionNetworkTglAddress[1], xArg, n ) ;
		break ;
	case XMEMO_SEARCH_RANGE :
		XtSetValues ( wDlgOptionNetworkTglAddress[2], xArg, n ) ;
		break ;
	}

	n = 0 ;
	XtSetArg ( xArg[n], XtNstring, pNetwork->pAddr[0] ) ; n ++ ;
	XtSetValues ( wDlgOptionNetworkTxtAddress[0], xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNstring, pNetwork->pAddr[1] ) ; n ++ ;
	XtSetValues ( wDlgOptionNetworkTxtAddress[1], xArg, n ) ;

	if ( pNetwork->bInsert ) {

		/* Remove from chained list */

		xmemo_RemoveFromChain ( &zNetworkAnchor, pNetwork->pChain ) ;

		/* Free allocated memory */

		xmemo_FreeNetworkEntry ( pNetwork ) ;

	} else {

		/* Delete pending until commit */

		pNetwork->bDelete = TRUE ;

	}

	/* Refresh list */

	xmemo_SetNetworkListArray ( ) ;

}

static XtCallbackRec pOptionNetworkDelete_CallbackRec [ ] =
{
	{ (XtCallbackProc) OptionNetworkDelete_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  OptionNetworkEnable_Click                                                 */
/******************************************************************************/

static void OptionNetworkEnable_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

zNetworkType		*pNetwork, *pChange ;

	if ( iListSelect < 0 || iListCount <= iListSelect ) return ;
	if ( bMemoTraceLog ) {
		char	sLog [ 50 ] ;
		sprintf ( sLog, "%d   %08X", iListSelect, pNetworkArray[iListSelect] ) ;
		xmemo_TraceLog ( "OptionNetworkEnable_Click", sLog ) ;
	}

	pNetwork = pNetworkArray[iListSelect] ;

	if ( pNetwork->bEnable ) return ;

	if ( pNetwork->bInsert ) {

		/* Change to enable state */

		pNetwork->bEnable = TRUE ;
		if ( pNetwork->pLine != NULL ) {
			XtFree ( (void*) pNetwork->pLine ) ;
			pNetwork->pLine = NULL ;
		}

	} else {

		/* Copy entry */

		pChange = xmemo_CopyNetworkEntry ( pNetwork ) ;
		if ( pChange == NULL ) return ;
		pChange->bEnable = TRUE ;
		pChange->bInsert = TRUE ;
		if ( pChange->pLine != NULL ) {
			XtFree ( (void*) pChange->pLine ) ;
			pChange->pLine = NULL ;
		}

		/* Insert into chained list */

		pChange->pChain = xmemo_InsertIntoChain ( &zNetworkAnchor, pChange ) ;
		if ( pChange->pChain == NULL ) {
			xmemo_ErrorLog ( "xmemo_CreateMemo", "xmemo_InsertIntoChain", NULL ) ;
			xmemo_FreeNetworkEntry ( pChange ) ;
			return ;
		}

		/* Delete pending until commit */

		pNetwork->bDelete = TRUE ;

	}

	/* Refresh list */

	xmemo_SetNetworkListArray ( ) ;

}

static XtCallbackRec pOptionNetworkEnable_CallbackRec [ ] =
{
	{ (XtCallbackProc) OptionNetworkEnable_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  OptionNetworkDisable_Click                                                */
/******************************************************************************/

static void OptionNetworkDisable_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

zNetworkType		*pNetwork, *pChange ;

	if ( iListSelect < 0 || iListCount <= iListSelect ) return ;
	if ( bMemoTraceLog ) {
		char	sLog [ 50 ] ;
		sprintf ( sLog, "%d   %08X", iListSelect, pNetworkArray[iListSelect] ) ;
		xmemo_TraceLog ( "OptionNetworkDisable_Click", sLog ) ;
	}

	pNetwork = pNetworkArray[iListSelect] ;

	if ( ! pNetwork->bEnable ) return ;

	if ( pNetwork->bInsert ) {

		/* Change to enable state */

		pNetwork->bEnable = FALSE ;
		if ( pNetwork->pLine != NULL ) {
			XtFree ( (void*) pNetwork->pLine ) ;
			pNetwork->pLine = NULL ;
		}

	} else {

		/* Copy entry */

		pChange = xmemo_CopyNetworkEntry ( pNetwork ) ;
		if ( pChange == NULL ) return ;
		pChange->bEnable = FALSE ;
		pChange->bInsert = TRUE ;
		if ( pChange->pLine != NULL ) {
			XtFree ( (void*) pChange->pLine ) ;
			pChange->pLine = NULL ;
		}

		/* Insert into chained list */

		pChange->pChain = xmemo_InsertIntoChain ( &zNetworkAnchor, pChange ) ;
		if ( pChange->pChain == NULL ) {
			xmemo_ErrorLog ( "xmemo_CreateMemo", "xmemo_InsertIntoChain", NULL ) ;
			xmemo_FreeNetworkEntry ( pChange ) ;
			return ;
		}

		/* Delete pending until commit */

		pNetwork->bDelete = TRUE ;

	}

	/* Refresh list */

	xmemo_SetNetworkListArray ( ) ;

}

static XtCallbackRec pOptionNetworkDisable_CallbackRec [ ] =
{
	{ (XtCallbackProc) OptionNetworkDisable_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  OptionNetworkClose_Click                                                  */
/******************************************************************************/

static void OptionNetworkClose_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

zChainType	*pChain ;
zNetworkType	*pNetwork ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "OptionNetworkClose_Click", NULL ) ;

	xmemo_PopdownDialog_OptionNetwork ( ) ;

	/* Commit search list */

	pChain = zNetworkAnchor.pHead ;
	while ( pChain != NULL ) {

		pNetwork = (zNetworkType*) ( pChain->pData ) ;
		pChain = pChain->pNext ;

		if ( pNetwork->bDelete ) {

			/* Notify to network */
			if ( pNetwork->bEnable  &&  pNetwork->bActive ) xmemo_NetNotify ( pNetwork, XMEMO_NET_NOTIFY_LEAVE ) ;

			/* Remove from chained list */
			xmemo_RemoveFromChain ( &zNetworkAnchor, pNetwork->pChain ) ;

			/* Free allocated memory */
			xmemo_FreeNetworkEntry ( pNetwork ) ;

		}

	}

	for ( pChain = zNetworkAnchor.pHead ; pChain != NULL ; pChain = pChain->pNext ) {

		pNetwork = (zNetworkType*) ( pChain->pData ) ;

		if ( pNetwork->bInsert ) {

			if ( pNetwork->bEnable ) xmemo_NetNotify ( pNetwork, XMEMO_NET_NOTIFY_ENTER ) ;

		}

	}

	/* Save */

	xmemo_SaveOptionNetwork ( ) ;

	for ( pChain = zNetworkAnchor.pHead ; pChain != NULL ; pChain = pChain->pNext ) {

		pNetwork = (zNetworkType*) ( pChain->pData ) ;
		pNetwork->bInsert = pNetwork->bDelete = FALSE ;

	}

}

static XtCallbackRec pOptionNetworkClose_CallbackRec [ ] =
{
	{ (XtCallbackProc) OptionNetworkClose_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  OptionNetworkCancel_Click                                                 */
/******************************************************************************/

static void OptionNetworkCancel_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

zChainType	*pChain ;
zNetworkType	*pNetwork ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "OptionNetworkCancel_Click", NULL ) ;

	xmemo_PopdownDialog_OptionNetwork ( ) ;

	/* Rollback search list */

	pChain = zNetworkAnchor.pHead ;
	while ( pChain != NULL ) {

		pNetwork = (zNetworkType*) ( pChain->pData ) ;
		pChain = pChain->pNext ;

		if ( pNetwork->bInsert ) {

			/* Remove from chained list */
			xmemo_RemoveFromChain ( &zNetworkAnchor, pNetwork->pChain ) ;

			/* Free allocated memory */
			xmemo_FreeNetworkEntry ( pNetwork ) ;

		}

	}

	for ( pChain = zNetworkAnchor.pHead ; pChain != NULL ; pChain = pChain->pNext ) {

		pNetwork = (zNetworkType*) ( pChain->pData ) ;
		pNetwork->bInsert = pNetwork->bDelete = FALSE ;

	}

}

static XtCallbackRec pOptionNetworkCancel_CallbackRec [ ] =
{
	{ (XtCallbackProc) OptionNetworkCancel_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  OptionNetworkList_Click                                                   */
/******************************************************************************/

static void OptionNetworkList_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

	iListSelect = ((XawListReturnStruct*)pCallData)->list_index ;

	if ( bMemoTraceLog ) {
		char	sLog [ 50 ] ;
		sprintf ( sLog, "%d   %08X", iListSelect, pNetworkArray[iListSelect] ) ;
		xmemo_TraceLog ( "OptionNetworkList_Click", sLog ) ;
	}

}

static XtCallbackRec pOptionNetworkList_CallbackRec [ ] =
{
	{ (XtCallbackProc) OptionNetworkList_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  xmemo_InitDialog_OptionNetwork                                            */
/******************************************************************************/

int xmemo_InitDialog_OptionNetwork ( )
{

int 	n ;
Arg 	xArg [ 16 ] ;
int 	w, h ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_InitDialog_OptionNetwork", NULL ) ;

	/* Dialog shell */

	n = 0 ;
	XtSetArg ( xArg[n], XtNtitle , "xmemo - Option [ Network ]" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNwidth , iDefaultSendDialogWidth  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNheight, iDefaultSendDialogHeight ) ; n ++ ;
	wDlgOptionNetwork = XtCreatePopupShell ( "DlgOptNetwork", transientShellWidgetClass, wTop, xArg, n ) ;
	                                  
	/* Main form */

	wDlgOptionNetworkForm = XtCreateManagedWidget ( "Form", formWidgetClass, wDlgOptionNetwork, xArg, 0 ) ;

	/* Edit area */

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth , 470 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop   , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbottom, XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft  , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright , XtChainRight  ) ; n ++ ;
	wDlgOptionNetworkEdit = XtCreateManagedWidget ( "Body", formWidgetClass, wDlgOptionNetworkForm, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNtop   , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbottom, XtChainBottom ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft  , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright , XtChainRight  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert, wDlgOptionNetworkEdit ) ; n ++ ;
	XtSetArg ( xArg[n], XtNallowHoriz, False ) ; n ++ ;
	XtSetArg ( xArg[n], XtNallowVert , True  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNuseBottom , True  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNuseRight  , True  ) ; n ++ ;
	wDlgOptionNetworkViewport = XtCreateManagedWidget ( "ListViewport", viewportWidgetClass, wDlgOptionNetworkForm, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth , 470 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNheight, 200 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop   , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbottom, XtChainBottom ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft  , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright , XtChainRight  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNdefaultColumns, 1     ) ; n ++ ;
	XtSetArg ( xArg[n], XtNforceColumns  , TRUE  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback, pOptionNetworkList_CallbackRec ) ; n ++ ;
	wDlgOptionNetworkList = XtCreateManagedWidget ( "List", listWidgetClass, wDlgOptionNetworkViewport, xArg, n ) ;

	/* Edit area ( Protocol ) */

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth    , 100 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNlabel   , "Protocol" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNjustify , XtJustifyLeft ) ; n ++ ;
	XtSetArg ( xArg[n], XtNborderWidth, 0 ) ; n ++ ;	
	XtSetArg ( xArg[n], XtNtop   , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft  , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright , XtChainLeft   ) ; n ++ ;
	wDlgOptionNetworkLblProtocol = XtCreateManagedWidget ( "Label", labelWidgetClass, wDlgOptionNetworkEdit, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth , 100 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNlabel , "IPMsg" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop   , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft  , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, wDlgOptionNetworkLblProtocol ) ; n ++ ;
	/*XtSetArg ( xArg[n], XtNradioGroup, wDlgOptionNetworkTglProtocol[1] ) ; n ++ ;*/
	/*XtSetArg ( xArg[n], XtNradioData, 0 ) ; n ++ ;*/
	wDlgOptionNetworkTglProtocol[0] = XtCreateManagedWidget ( "TglButton", toggleWidgetClass, wDlgOptionNetworkEdit, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth , 100 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNlabel , "Oboegaki" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop   , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft  , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, wDlgOptionNetworkTglProtocol[0] ) ; n ++ ;
	/*XtSetArg ( xArg[n], XtNradioGroup, wDlgOptionNetworkTglProtocol[0] ) ; n ++ ;*/	
	/*XtSetArg ( xArg[n], XtNradioData, 1 ) ; n ++ ;*/
	wDlgOptionNetworkTglProtocol[1] = XtCreateManagedWidget ( "TglButton", toggleWidgetClass, wDlgOptionNetworkEdit, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth , 100 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNlabel , "Husenshi" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop   , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft  , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, wDlgOptionNetworkTglProtocol[1] ) ; n ++ ;
#if 1
	XtSetArg ( xArg[n], XtNmappedWhenManaged, False ) ; n ++ ;
#endif
	/*XtSetArg ( xArg[n], XtNradioGroup, wDlgOptionNetworkTglProtocol[1] ) ; n ++ ;*/
	/*XtSetArg ( xArg[n], XtNradioData, 2 ) ; n ++ ;*/
	wDlgOptionNetworkTglProtocol[2] = XtCreateManagedWidget ( "TglButton", toggleWidgetClass, wDlgOptionNetworkEdit, xArg, n ) ;

	XawToggleChangeRadioGroup ( wDlgOptionNetworkTglProtocol[0], wDlgOptionNetworkTglProtocol[2] ) ;
	XawToggleChangeRadioGroup ( wDlgOptionNetworkTglProtocol[1], wDlgOptionNetworkTglProtocol[0] ) ;
	XawToggleChangeRadioGroup ( wDlgOptionNetworkTglProtocol[2], wDlgOptionNetworkTglProtocol[1] ) ;

	/* Edit Area ( Port ) */

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth    , 100 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNlabel   , "Port" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNjustify , XtJustifyLeft ) ; n ++ ;
	XtSetArg ( xArg[n], XtNborderWidth, 0 ) ; n ++ ;	
	XtSetArg ( xArg[n], XtNtop   , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft  , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert, wDlgOptionNetworkLblProtocol ) ; n ++ ;
	wDlgOptionNetworkLblPort = XtCreateManagedWidget ( "Label", labelWidgetClass, wDlgOptionNetworkEdit, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth    , 50 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop      , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft     , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright    , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert , wDlgOptionNetworkLblProtocol ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, wDlgOptionNetworkLblPort ) ; n ++ ;	
	XtSetArg ( xArg[n], XtNeditType , XawtextEdit   ) ; n ++ ;	
	wDlgOptionNetworkTxtPort = XtCreateManagedWidget ( "Text", asciiTextWidgetClass, wDlgOptionNetworkEdit, xArg, n ) ;

	/* Edit Area ( Address ) */

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth    , 100 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNlabel   , "Address" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNjustify , XtJustifyLeft ) ; n ++ ;
	XtSetArg ( xArg[n], XtNborderWidth, 0 ) ; n ++ ;	
	XtSetArg ( xArg[n], XtNtop   , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft  , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert, wDlgOptionNetworkLblPort ) ; n ++ ;
	wDlgOptionNetworkLblAddress[0] = XtCreateManagedWidget ( "Label", labelWidgetClass, wDlgOptionNetworkEdit, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth    , 100 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNlabel    , "Broadcast" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop      , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft     , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright    , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert , wDlgOptionNetworkLblPort ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, wDlgOptionNetworkLblAddress[0] ) ; n ++ ;
	/*XtSetArg ( xArg[n], XtNradioGroup, wDlgOptionNetworkTglAddress[2] ) ; n ++ ;*/
	wDlgOptionNetworkTglAddress[0] = XtCreateManagedWidget ( "TglButton", toggleWidgetClass, wDlgOptionNetworkEdit, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth    , 100 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNlabel    , "Host" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop      , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft     , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright    , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert , wDlgOptionNetworkLblPort  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, wDlgOptionNetworkTglAddress[0] ) ; n ++ ;
	/*XtSetArg ( xArg[n], XtNradioGroup, wDlgOptionNetworkTglAddress[0] ) ; n ++ ;*/	
	wDlgOptionNetworkTglAddress[1] = XtCreateManagedWidget ( "TglButton", toggleWidgetClass, wDlgOptionNetworkEdit, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth    , 100 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNlabel    , "Range" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop      , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft     , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright    , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert , wDlgOptionNetworkLblPort  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, wDlgOptionNetworkTglAddress[1] ) ; n ++ ;
	/*XtSetArg ( xArg[n], XtNradioGroup, wDlgOptionNetworkTglAddress[1] ) ; n ++ ;*/
	wDlgOptionNetworkTglAddress[2] = XtCreateManagedWidget ( "TglButton", toggleWidgetClass, wDlgOptionNetworkEdit, xArg, n ) ;

	XawToggleChangeRadioGroup ( wDlgOptionNetworkTglAddress[0], wDlgOptionNetworkTglAddress[2] ) ;
	XawToggleChangeRadioGroup ( wDlgOptionNetworkTglAddress[1], wDlgOptionNetworkTglAddress[0] ) ;
	XawToggleChangeRadioGroup ( wDlgOptionNetworkTglAddress[2], wDlgOptionNetworkTglAddress[1] ) ;
	XawToggleUnsetCurrent ( wDlgOptionNetworkTglAddress[0] ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth    , 150 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop      , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft     , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright    , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert , wDlgOptionNetworkLblAddress[0] ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, wDlgOptionNetworkLblAddress[0] ) ; n ++ ;	
	XtSetArg ( xArg[n], XtNeditType , XawtextEdit   ) ; n ++ ;	
	wDlgOptionNetworkTxtAddress[0] = XtCreateManagedWidget ( "Text", asciiTextWidgetClass, wDlgOptionNetworkEdit, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel    , "  -  " ) ; n ++ ;
	XtSetArg ( xArg[n], XtNjustify  , XtJustifyCenter ) ; n ++ ;
	XtSetArg ( xArg[n], XtNborderWidth, 0 ) ; n ++ ;	
	XtSetArg ( xArg[n], XtNtop      , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft     , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright    , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert , wDlgOptionNetworkLblAddress[0] ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, wDlgOptionNetworkTxtAddress[0] ) ; n ++ ;	
	wDlgOptionNetworkLblAddress[1] = XtCreateManagedWidget ( "Label", labelWidgetClass, wDlgOptionNetworkEdit, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth    , 150 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop      , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft     , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright    , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert , wDlgOptionNetworkLblAddress[0] ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, wDlgOptionNetworkLblAddress[1] ) ; n ++ ;
	XtSetArg ( xArg[n], XtNeditType , XawtextEdit   ) ; n ++ ;	
	wDlgOptionNetworkTxtAddress[1] = XtCreateManagedWidget ( "Text", asciiTextWidgetClass, wDlgOptionNetworkEdit, xArg, n ) ;

	/* Command button */

	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel    , "Add"         ) ; n ++ ;
	XtSetArg ( xArg[n], XtNwidth    , 70            ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop      , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbottom   , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft     , XtChainRight  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright    , XtChainRight  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, wDlgOptionNetworkEdit ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback , pOptionNetworkAdd_CallbackRec ) ; n ++ ;
	XtCreateManagedWidget ( "CmdButton", commandWidgetClass, wDlgOptionNetworkForm, xArg, n ) ;

	/* Command button ( Right ) */

	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel    , "Delete"      ) ; n ++ ;
	XtSetArg ( xArg[n], XtNwidth    , 70            ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop      , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbottom   , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft     , XtChainRight  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright    , XtChainRight  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert , wDlgOptionNetworkEdit ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, wDlgOptionNetworkViewport ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback , pOptionNetworkDelete_CallbackRec ) ; n ++ ;
	wDlgOptionNetworkCmdDelete = XtCreateManagedWidget ( "CmdButton", commandWidgetClass, wDlgOptionNetworkForm, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel    , "Enable"      ) ; n ++ ;
	XtSetArg ( xArg[n], XtNwidth    , 70            ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop      , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbottom   , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft     , XtChainRight  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright    , XtChainRight  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert , wDlgOptionNetworkCmdDelete ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, wDlgOptionNetworkViewport  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback , pOptionNetworkEnable_CallbackRec ) ; n ++ ;
	wDlgOptionNetworkCmdEnable = XtCreateManagedWidget ( "CmdButton", commandWidgetClass, wDlgOptionNetworkForm, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel    , "Disable"     ) ; n ++ ;
	XtSetArg ( xArg[n], XtNwidth    , 70            ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop      , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbottom   , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft     , XtChainRight  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright    , XtChainRight  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert , wDlgOptionNetworkCmdEnable ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, wDlgOptionNetworkViewport  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback , pOptionNetworkDisable_CallbackRec ) ; n ++ ;
	wDlgOptionNetworkCmdDisable = XtCreateManagedWidget ( "CmdButton", commandWidgetClass, wDlgOptionNetworkForm, xArg, n ) ;

	/* Command button ( Bottom ) */

	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel    , "OK" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop      , XtChainBottom ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbottom   , XtChainBottom ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft     , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright    , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert , wDlgOptionNetworkViewport ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback , pOptionNetworkClose_CallbackRec ) ; n ++ ;
	wDlgOptionNetworkCmdClose = XtCreateManagedWidget ( "CmdButton", commandWidgetClass, wDlgOptionNetworkForm, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel    , "Cancel" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop      , XtChainBottom ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbottom   , XtChainBottom ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft     , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright    , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert , wDlgOptionNetworkViewport ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, wDlgOptionNetworkCmdClose ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback , pOptionNetworkCancel_CallbackRec ) ; n ++ ;
	wDlgOptionNetworkCmdCancel = XtCreateManagedWidget ( "CmdButton", commandWidgetClass, wDlgOptionNetworkForm, xArg, n ) ;

	return RC_SUCCESS ;

}

/******************************************************************************/
/*  xmemo_PopupDialog_OptionNetwork                                           */
/******************************************************************************/

int xmemo_PopupDialog_OptionNetwork ( )
{

int 	n ;
Arg 	xArg [ 3 ] ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_PopupDialog_OptionNetwork", NULL ) ;

	xmemo_SetNetworkListArray ( ) ;

	XtPopup ( wDlgOptionNetwork, XtGrabExclusive ) ;

	return RC_SUCCESS ;
	
}

/******************************************************************************/
/*  xmemo_PopdownDialog_OptionNetwork                                         */
/******************************************************************************/

int xmemo_PopdownDialog_OptionNetwork ( )
{

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_PopdownDialog_OptionNetwork", NULL ) ;

	XtPopdown ( wDlgOptionNetwork ) ;

	return RC_SUCCESS ;

}

/******************************************************************************/
/*  xmemo_FreeNetworkAll                                                      */
/******************************************************************************/

int xmemo_FreeNetworkAll ( )
{

zNetworkType		*pNetwork ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_FreeNetworkAll", NULL ) ;

	while ( zNetworkAnchor.pHead != NULL ) {

		pNetwork = (zNetworkType*) ( zNetworkAnchor.pHead->pData ) ;

		/* Remove from chained list */

		xmemo_RemoveFromChain ( &zNetworkAnchor, zNetworkAnchor.pHead ) ;

		/* Free allocated memory */

		xmemo_FreeNetworkEntry ( pNetwork ) ;

	}
	return RC_SUCCESS ;

}

