/******************************************************************************/
/*                                                                            */
/*  <<< Copyright >>>                                                         */
/*                                                                            */
/*  The "xmemo" software is copyright 1999 by Takao Abe.                      */
/*  All rights reserved.                                                      */
/*                                                                            */
/*  The "xmemo" software refers to the source code, the binary code,          */
/*  the documents, and all derivatives based on this software under           */
/*  the copyright law, that is a software containing a program or a           */
/*  portion of a program, regardless of modifications and/or translations.    */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  <<< Licence and redistribution >>>                                        */
/*                                                                            */
/*  The source code and the binary code are free software.                    */
/*  The use and redistribution of source code and binary code are permitted   */
/*  provided that the following conditions are met:                           */
/*  You may use the software for private use and commercial use,              */
/*  and may modify your copy or copies of the software or any portion of it.  */
/*                                                                            */
/*  THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESSED OR IMPLIED            */
/*  WARRANTIES OF ANY KIND, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED        */
/*  WARRANTIES OF MERCHANTABLILITY AND FITNESS FOR A PARTICULAR PURPOSE.      */
/*  IN NO EVENT SHALL THE AUTHOR TAKAO ABE BE LIABLE FOR ANY DIRECT,          */
/*  INDIRECT, GENERAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES           */
/*  ( INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR       */
/*  SERVICES; LOSS OF USE, DATA OR PROFITS; OR BUSINESS INTERRUPTION )        */
/*  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,       */
/*  STRICT LIABILITY, OR TORT ( INCLUDING NEGLIGENCE OR OTHERWISE )           */
/*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED       */
/*  OF THE POSSIBILITY OF SUCH DAMAGE.                                        */
/*                                                                            */
/*  Redistribution and use in source form or binary form, with or without     */
/*  modification, the whole program or a portion of it, are permitted         */
/*  provided that one retains the entire copyright notice properly, and       */
/*  displays or retains this license in the documentation and/or other        */
/*  materials provided with the distribution.                                 */
/*                                                                            */
/*  Redistribution and use in source form or binary form, with modification,  */
/*  the whole program or a portion of it, are permitted provided that one     */
/*  declares that it is modified software and not original, and displays or   */
/*  describes any modifications in the documentation and/or other materials   */
/*  provided with the distribution.                                           */
/*                                                                            */
/*  This software and the name of the author must not be used to endorse or   */
/*  promote products derived from this software without prior written         */
/*  permission.                                                               */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  EMail   abetakao@bea.hi-ho.ne.jp                                          */
/*          HCA00763@nifty.ne.jp                                              */
/*                                                                            */
/*  URL     http://www.bea.hi-ho.ne.jp/abetakao/                              */
/*                                                                            */
/******************************************************************************/


#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<time.h>


#include	<X11/Xlib.h>
#include	<X11/Xutil.h>
#include	<X11/Intrinsic.h>
#include	<X11/StringDefs.h>
#include	<X11/Shell.h>
#include	<X11/cursorfont.h>

#include	<X11/Xaw/AsciiText.h>
#include	<X11/Xaw/Box.h>
#include	<X11/Xaw/Form.h>
#include	<X11/Xaw/List.h>
#include	<X11/Xaw/MenuButton.h>
#include	<X11/Xaw/Simple.h>
#include	<X11/Xaw/SimpleMenu.h>
#include	<X11/Xaw/SmeBSB.h>
#include	<X11/Xaw/SmeLine.h>
#include	<X11/Xaw/Viewport.h>


#include	"xmemo.h"
#include	"xmemodata.h"
#include	"xmemofunc.h"


/******************************************************************************/
/*  Local variables                                                           */
/******************************************************************************/


static	Widget	wTopMenuButton [ 4 ] ;
static	Widget	wTopPulldownMenu [ 4 ] ;
static	Widget	wTopPulldownMenuPresence ;
static	Widget	wTopPulldownMenuAbsence ;

static	int 	iListSize = 100 ;
static	int 	iListCount ;
static	char	**pListArray ;


/******************************************************************************/
/*  xmemo_StatusMsg                                                           */
/******************************************************************************/

void xmemo_StatusMsg ( char *pMsg )
{

int 	i, n ;
Arg 	xArg [ 4 ] ;

	if ( iListCount < iListSize ) {
		iListCount ++ ;
	} else {
		if ( pListArray[0] != NULL ) XtFree ( pListArray[0] ) ;
		for ( i = 0 ; i < iListSize - 1 ; i ++ ) pListArray[i] = pListArray[i+1] ;
	}

	pListArray[iListCount-1] = XtMalloc ( strlen(pMsg) + 1 ) ;
	if ( pListArray[iListCount-1] == NULL ) {
		if ( iListCount > 0 ) iListCount -- ;
		return ;
	}

	strcpy ( pListArray[iListCount-1], pMsg ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNlist, pListArray ) ; n ++ ;
	XtSetArg ( xArg[n], XtNnumberStrings, iListCount ) ; n ++ ;
	XtSetValues ( wTopStatusList, xArg, n ) ;
	XawListChange ( wTopStatusList, pListArray, iListCount, 0, False ) ;

}

/******************************************************************************/
/*  MenuMemo_Click                                                            */
/******************************************************************************/

static void MenuMemo_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

zMemoType		*pMemo ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "MenuMemo_Click", NULL ) ;

	pMemoActive = xmemo_CreateMemo ( ) ;
	if ( pMemoActive == NULL ) return ;
	xmemo_PopupDialog_Edit ( ) ;

}

static XtCallbackRec pMenuMemo_CallbackRec [ ] =
{
	{ (XtCallbackProc) MenuMemo_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  MenuPresence_Click                                                        */
/******************************************************************************/

static void MenuPresence_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

int 	n ;
Arg 	xArg [ 2 ] ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "MenuPresence_Click", NULL ) ;

	if ( ! bUserAbsence ) return ;

	bUserAbsence = FALSE ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNsensitive, False ) ; n ++ ;
	XtSetValues ( wTopPulldownMenuPresence, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNsensitive, True ) ; n ++ ;
	XtSetValues ( wTopPulldownMenuAbsence, xArg, n ) ;

	xmemo_NetNotify_Broadcast ( XMEMO_NET_NOTIFY_PRESENCE ) ;

}

static XtCallbackRec pMenuPresence_CallbackRec [ ] =
{
	{ (XtCallbackProc) MenuPresence_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  MenuAbsence_Click                                                        */
/******************************************************************************/

static void MenuAbsence_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

int 	n ;
Arg 	xArg [ 2 ] ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "MenuAbsence_Click", NULL ) ;

	if ( bUserAbsence ) return ;

	bUserAbsence = TRUE ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNsensitive, True ) ; n ++ ;
	XtSetValues ( wTopPulldownMenuPresence, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNsensitive, False ) ; n ++ ;
	XtSetValues ( wTopPulldownMenuAbsence, xArg, n ) ;

	xmemo_NetNotify_Broadcast ( XMEMO_NET_NOTIFY_ABSENCE ) ;

}

static XtCallbackRec pMenuAbsence_CallbackRec [ ] =
{
	{ (XtCallbackProc) MenuAbsence_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  MenuQuit_Click                                                            */
/******************************************************************************/

static void MenuQuit_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

	if ( bMemoTraceLog ) xmemo_TraceLog ( "MenuQuit_Click", NULL ) ;

	xmemo_Exit ( ) ;

	exit ( 0 ) ;

}

static XtCallbackRec pMenuQuit_CallbackRec [ ] =
{
	{ (XtCallbackProc) MenuQuit_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  MenuOptionPreference_Click                                                */
/******************************************************************************/

static void MenuOptionPreference_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

	if ( bMemoTraceLog ) xmemo_TraceLog ( "MenuOptionPreference_Click", NULL ) ;

	xmemo_PopupDialog_OptionPreference ( ) ;

}

static XtCallbackRec pMenuOptionPreference_CallbackRec [ ] =
{
	{ (XtCallbackProc) MenuOptionPreference_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  MenuOptionProfile_Click                                                   */
/******************************************************************************/

static void MenuOptionProfile_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

	if ( bMemoTraceLog ) xmemo_TraceLog ( "MenuOptionProfile_Click", NULL ) ;

	xmemo_PopupDialog_OptionProfile ( ) ;

}

static XtCallbackRec pMenuOptionProfile_CallbackRec [ ] =
{
	{ (XtCallbackProc) MenuOptionProfile_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  MenuOptionNetwork_Click                                                   */
/******************************************************************************/

static void MenuOptionNetwork_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

	if ( bMemoTraceLog ) xmemo_TraceLog ( "MenuOptionNetwork_Click", NULL ) ;

	xmemo_PopupDialog_OptionNetwork ( ) ;

}

static XtCallbackRec pMenuOptionNetwork_CallbackRec [ ] =
{
	{ (XtCallbackProc) MenuOptionNetwork_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  MenuOptionLog_Click                                                       */
/******************************************************************************/

static void MenuOptionLog_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

	if ( bMemoTraceLog ) xmemo_TraceLog ( "MenuOptionLog_Click", NULL ) ;

	xmemo_PopupDialog_OptionLog ( ) ;

}

static XtCallbackRec pMenuOptionLog_CallbackRec [ ] =
{
	{ (XtCallbackProc) MenuOptionLog_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  MenuOptionSystem_Click                                                    */
/******************************************************************************/

static void MenuOptionSystem_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

	if ( bMemoTraceLog ) xmemo_TraceLog ( "MenuOptionSystem_Click", NULL ) ;

	xmemo_PopupDialog_OptionSystem ( ) ;

}

static XtCallbackRec pMenuOptionSystem_CallbackRec [ ] =
{
	{ (XtCallbackProc) MenuOptionSystem_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  MenuWindowRaise_Click                                                     */
/******************************************************************************/

static void MenuWindowRaise_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

zMemoType		*pMemo ;
zChainType		*pChain ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "MenuWindowRaise_Click", NULL ) ;

	for ( pChain = zMemoAnchor.pHead ; pChain != NULL ; pChain = pChain->pNext ) {

		pMemo = (zMemoType*) ( pChain->pData ) ;
		XRaiseWindow ( XtDisplay(pMemo->wMemo), XtWindow(pMemo->wMemo) ) ;

	}

}

static XtCallbackRec pMenuWindowRaise_CallbackRec [ ] =
{
	{ (XtCallbackProc) MenuWindowRaise_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  MenuWindowLower_Click                                                     */
/******************************************************************************/

static void MenuWindowLower_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

zMemoType		*pMemo ;
zChainType		*pChain ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "MenuWindowLower_Click", NULL ) ;

	for ( pChain = zMemoAnchor.pTail ; pChain != NULL ; pChain = pChain->pBack ) {

		pMemo = (zMemoType*) ( pChain->pData ) ;
		XLowerWindow ( XtDisplay(pMemo->wMemo), XtWindow(pMemo->wMemo) ) ;

	}

}

static XtCallbackRec pMenuWindowLower_CallbackRec [ ] =
{
	{ (XtCallbackProc) MenuWindowLower_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  MenuHelpAbout_Click                                                       */
/******************************************************************************/

static void MenuHelpAbout_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

zMemoType		*pMemo ;
zChainType		*pChain ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "MenuHelpAbout_Click", NULL ) ;

	xmemo_PopupDialog_Aboutxmemo ( ) ;

}

static XtCallbackRec pMenuHelpAbout_CallbackRec [ ] =
{
	{ (XtCallbackProc) MenuHelpAbout_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  xmemo_InitWindow                                                          */
/******************************************************************************/

int xmemo_InitWindow ( int argc, char *argv[ ] )
{

static	Arg	xArgForm [ ] = {
		{ XtNtop      , XtChainTop },
		{ XtNbottom   , XtChainTop },
		{ XtNresizable, True       }
} ;

zMemoType		*pMemo ;
zChainType		*pChain ;
int 	i, rc ;
int 	n ;
Arg 	xArg [ 16 ] ;

Widget	wTopMenuMemo, wTopMenuOption, wTopMenuHelp ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_InitWindow", NULL ) ;

	pListArray = (char**) XtCalloc ( sizeof(char*), iListSize + 1 ) ;

	/* Main app */

	XtSetLanguageProc ( NULL, NULL, NULL ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth , 440 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNheight, 200 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop   , XtChainTop ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft  , XtChainLeft ) ; n ++ ;

	wTop = XtAppInitialize ( &xApp, "XMemo", NULL, 0, &argc, argv, NULL, xArg, n ) ;

	iScreenWidth  = XDisplayWidth  ( XtDisplay(wTop), 0 ) ;
	iScreenHeight = XDisplayHeight ( XtDisplay(wTop), 0 ) ;

	/* Check command line parameter */

	rc = xmemo_Param ( argc, argv ) ;
	if ( rc == RC_ERROR ) {
		printf ( "xmemo : Invalid parameter\n" ) ;
		exit ( 1 ) ;
	}

	/* Main form */

	wTopForm = XtCreateManagedWidget ( "Form", formWidgetClass, wTop, xArgForm, 3 ) ;

	/* Menu area */

	n = 0 ;
	XtSetArg ( xArg[n], XtNborderWidth, 0      ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop   , XtChainTop  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbottom, XtChainTop  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft  , XtChainLeft ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright , XtChainLeft ) ; n ++ ;
	wTopMenu = XtCreateManagedWidget ( "MenuBar", formWidgetClass, wTopForm, xArg, n ) ;

	/* Menu button */

	n = 0 ;
	XtSetArg ( xArg[n], XtNmenuName , "PulldownMenuMemo" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNlabel    , "Memo" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft     , XtChainLeft ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright    , XtChainLeft ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, NULL ) ; n ++ ;
	wTopMenuButton[0] = XtCreateManagedWidget ( "MenuButton", menuButtonWidgetClass, wTopMenu, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNmenuName , "PulldownMenuOptions" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNlabel    , "Options" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft     , XtChainLeft ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright    , XtChainLeft ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, wTopMenuButton[0] ) ; n ++ ;
	wTopMenuButton[1] = XtCreateManagedWidget ( "MenuButton", menuButtonWidgetClass, wTopMenu, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNmenuName , "PulldownMenuWindow" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNlabel    , "Window" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft     , XtChainLeft ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright    , XtChainLeft ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, wTopMenuButton[1] ) ; n ++ ;
	wTopMenuButton[2] = XtCreateManagedWidget ( "MenuButton", menuButtonWidgetClass, wTopMenu, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNmenuName , "PulldownMenuHelp" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNlabel    , "Help" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft     , XtChainLeft ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright    , XtChainLeft ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, wTopMenuButton[2] ) ; n ++ ;
	wTopMenuButton[3] = XtCreateManagedWidget ( "MenuButton", menuButtonWidgetClass, wTopMenu, xArg, n ) ;

	/* Status log */

	n = 0 ;
	XtSetArg ( xArg[n], XtNtop   , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbottom, XtChainBottom ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft  , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright , XtChainRight  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert  , wTopMenu ) ; n ++ ;
	XtSetArg ( xArg[n], XtNallowHoriz, True  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNallowVert , True  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNuseBottom , True  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNuseRight  , True  ) ; n ++ ;
	wTopStatusView = XtCreateManagedWidget ( "ListViewport", viewportWidgetClass, wTopForm, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNtop   , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbottom, XtChainBottom ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft  , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright , XtChainRight  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNdefaultColumns, 1     ) ; n ++ ;
	XtSetArg ( xArg[n], XtNforceColumns  , TRUE  ) ; n ++ ;
	wTopStatusList = XtCreateManagedWidget ( "List", listWidgetClass, wTopStatusView, xArg, n ) ;

	xmemo_StatusMsg ( pVersion ) ;
	xmemo_StatusMsg ( pCopyright ) ;
	xmemo_StatusMsg ( "" ) ;

	/* Pulldown menu */

	wTopPulldownMenu[0] = XtCreatePopupShell ( "PulldownMenuMemo", simpleMenuWidgetClass, wTopMenuButton[0], NULL, 0 ) ;
	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel, "New memo" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback, pMenuMemo_CallbackRec ) ; n ++ ;
	XtCreateManagedWidget ( "MenuCmd", smeBSBObjectClass, wTopPulldownMenu[0], xArg, n ) ;
	XtCreateManagedWidget ( "MenuLine", smeLineObjectClass, wTopPulldownMenu[0], NULL, 0 ) ;
	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel, "Absence -> Presence" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNsensitive, False ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback, pMenuPresence_CallbackRec ) ; n ++ ;
	wTopPulldownMenuPresence = XtCreateManagedWidget ( "MenuCmd", smeBSBObjectClass, wTopPulldownMenu[0], xArg, n ) ;
	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel, "Presence -> Absence" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback, pMenuAbsence_CallbackRec ) ; n ++ ;
	wTopPulldownMenuAbsence = XtCreateManagedWidget ( "MenuCmd", smeBSBObjectClass, wTopPulldownMenu[0], xArg, n ) ;
	XtCreateManagedWidget ( "MenuLine", smeLineObjectClass, wTopPulldownMenu[0], NULL, 0 ) ;
	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel, "Quit" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback, pMenuQuit_CallbackRec ) ; n ++ ;
	XtCreateManagedWidget ( "MenuCmd", smeBSBObjectClass, wTopPulldownMenu[0], xArg, n ) ;

	wTopPulldownMenu[1] = XtCreatePopupShell ( "PulldownMenuOptions", simpleMenuWidgetClass, wTopMenuButton[1], NULL, 0  ) ;
	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel, "Preference" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback, pMenuOptionPreference_CallbackRec ) ; n ++ ;
	XtCreateManagedWidget ( "MenuCmd", smeBSBObjectClass, wTopPulldownMenu[1], xArg, n ) ;
	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel, "Profile" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback, pMenuOptionProfile_CallbackRec ) ; n ++ ;
	XtCreateManagedWidget ( "MenuCmd", smeBSBObjectClass, wTopPulldownMenu[1], xArg, n ) ;
	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel, "Network" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback, pMenuOptionNetwork_CallbackRec ) ; n ++ ;
	XtCreateManagedWidget ( "MenuCmd", smeBSBObjectClass, wTopPulldownMenu[1], xArg, n ) ;
#if 0
	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel, "Log" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback, pMenuOptionLog_CallbackRec ) ; n ++ ;
	XtCreateManagedWidget ( "MenuCmd", smeBSBObjectClass, wTopPulldownMenu[1], xArg, n ) ;
#endif
	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel, "System" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback, pMenuOptionSystem_CallbackRec ) ; n ++ ;
	XtCreateManagedWidget ( "MenuCmd", smeBSBObjectClass, wTopPulldownMenu[1], xArg, n ) ;

	wTopPulldownMenu[2] = XtCreatePopupShell ( "PulldownMenuWindow", simpleMenuWidgetClass, wTopMenuButton[2], NULL, 0  ) ;
	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel, "Raise Z order" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback, pMenuWindowRaise_CallbackRec ) ; n ++ ;
	XtCreateManagedWidget ( "MenuCmd", smeBSBObjectClass, wTopPulldownMenu[2], xArg, n ) ;
	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel, "Lower Z order" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback, pMenuWindowLower_CallbackRec ) ; n ++ ;
	XtCreateManagedWidget ( "MenuCmd", smeBSBObjectClass, wTopPulldownMenu[2], xArg, n ) ;

	wTopPulldownMenu[3] = XtCreatePopupShell ( "PulldownMenuHelp", simpleMenuWidgetClass, wTopMenuButton[3], NULL, 0  ) ;
/*
	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel, "Help" ) ; n ++ ;
	XtCreateManagedWidget ( "MenuCmd", smeBSBObjectClass, wTopPulldownMenu[3], xArg, n ) ;
*/
	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel, "About xmemo" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback, pMenuHelpAbout_CallbackRec ) ; n ++ ;
	XtCreateManagedWidget ( "MenuCmd", smeBSBObjectClass, wTopPulldownMenu[3], xArg, n ) ;

	/* Realize top widget */

	XtRealizeWidget ( wTop ) ;

	/* Memo widget */

	xmemo_InitWindow_Memo ( ) ;

	/* Dialogs */

	xmemo_InitDialog_Edit ( ) ;

	xmemo_InitDialog_Font ( ) ;

	xmemo_InitDialog_Send ( ) ;

	xmemo_InitDialog_OptionPreference ( ) ;

	xmemo_InitDialog_OptionProfile ( ) ;

	xmemo_InitDialog_OptionNetwork ( ) ;

	xmemo_InitDialog_OptionLog ( ) ;

	xmemo_InitDialog_OptionSystem ( ) ;

	xmemo_InitDialog_Aboutxmemo ( ) ;

	/* Memo */

	for ( pChain = zMemoAnchor.pHead ; pChain != NULL ; pChain = pChain->pNext ) {

		xmemo_CreateMemoWidget ( (zMemoType*) ( pChain->pData ) ) ;

	}

	return RC_SUCCESS ;

}

/******************************************************************************/
/*  xmemo_ExitWindow                                                          */
/******************************************************************************/

int xmemo_ExitWindow ( )
{

	return RC_SUCCESS ;

}
