/******************************************************************************/
/*                                                                            */
/*  <<< Copyright >>>                                                         */
/*                                                                            */
/*  The "xmemo" software is copyright 1999 by Takao Abe.                      */
/*  All rights reserved.                                                      */
/*                                                                            */
/*  The "xmemo" software refers to the source code, the binary code,          */
/*  the documents, and all derivatives based on this software under           */
/*  the copyright law, that is a software containing a program or a           */
/*  portion of a program, regardless of modifications and/or translations.    */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  <<< Licence and redistribution >>>                                        */
/*                                                                            */
/*  The source code and the binary code are free software.                    */
/*  The use and redistribution of source code and binary code are permitted   */
/*  provided that the following conditions are met:                           */
/*  You may use the software for private use and commercial use,              */
/*  and may modify your copy or copies of the software or any portion of it.  */
/*                                                                            */
/*  THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESSED OR IMPLIED            */
/*  WARRANTIES OF ANY KIND, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED        */
/*  WARRANTIES OF MERCHANTABLILITY AND FITNESS FOR A PARTICULAR PURPOSE.      */
/*  IN NO EVENT SHALL THE AUTHOR TAKAO ABE BE LIABLE FOR ANY DIRECT,          */
/*  INDIRECT, GENERAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES           */
/*  ( INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR       */
/*  SERVICES; LOSS OF USE, DATA OR PROFITS; OR BUSINESS INTERRUPTION )        */
/*  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,       */
/*  STRICT LIABILITY, OR TORT ( INCLUDING NEGLIGENCE OR OTHERWISE )           */
/*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED       */
/*  OF THE POSSIBILITY OF SUCH DAMAGE.                                        */
/*                                                                            */
/*  Redistribution and use in source form or binary form, with or without     */
/*  modification, the whole program or a portion of it, are permitted         */
/*  provided that one retains the entire copyright notice properly, and       */
/*  displays or retains this license in the documentation and/or other        */
/*  materials provided with the distribution.                                 */
/*                                                                            */
/*  Redistribution and use in source form or binary form, with modification,  */
/*  the whole program or a portion of it, are permitted provided that one     */
/*  declares that it is modified software and not original, and displays or   */
/*  describes any modifications in the documentation and/or other materials   */
/*  provided with the distribution.                                           */
/*                                                                            */
/*  This software and the name of the author must not be used to endorse or   */
/*  promote products derived from this software without prior written         */
/*  permission.                                                               */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  EMail   abetakao@bea.hi-ho.ne.jp                                          */
/*          HCA00763@nifty.ne.jp                                              */
/*                                                                            */
/*  URL     http://www.bea.hi-ho.ne.jp/abetakao/                              */
/*                                                                            */
/******************************************************************************/


#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<time.h>

#include	<X11/Intrinsic.h>

#include	"xmemo.h"
#include	"xmemodata.h"
#include	"xmemofunc.h"


/******************************************************************************/
/*  xmemo_FreeMemo                                                            */
/******************************************************************************/

void xmemo_FreeMemo ( zMemoType *pMemo )
{

	if ( pMemo->pFontStruct != NULL ) XFreeFont ( XtDisplay(wTop), pMemo->pFontStruct ) ;
	if ( pMemo->pFontName != NULL ) XtFree ( (void*) ( pMemo->pFontName ) ) ;
	if ( pMemo->pHost     != NULL ) XtFree ( (void*) ( pMemo->pHost     ) ) ;
	if ( pMemo->pUser     != NULL ) XtFree ( (void*) ( pMemo->pUser     ) ) ;
	if ( pMemo->pNick     != NULL ) XtFree ( (void*) ( pMemo->pNick     ) ) ;
	if ( pMemo->pGroup    != NULL ) XtFree ( (void*) ( pMemo->pGroup    ) ) ;
	if ( pMemo->pText     != NULL ) XtFree ( (void*) ( pMemo->pText     ) ) ;
	XtFree ( (void*) pMemo ) ;

}

/******************************************************************************/
/*  xmemo_LoadMemo_Token                                                      */
/******************************************************************************/

static char* xmemo_LoadMemo_Token ( char *pBuf, char *pDelimiter, char *pValidation )
{

char	*pToken ;

	pToken = strtok ( pBuf, pDelimiter ) ;
	if ( pToken == NULL ) return NULL ;
	if ( strlen ( pToken ) == 0 ) return NULL ;
	if ( strspn ( pToken, pValidation ) != strlen ( pToken ) ) return NULL ;

	return pToken ;

}

/******************************************************************************/
/*  xmemo_LoadMemo_FormatVersion_1                                            */
/******************************************************************************/

static zMemoType* xmemo_LoadMemo_FormatVersion_1 ( char *pBuf )
{

char	*pLine [ 32 ] ;
char	*pToken [ 16 ] ;
int 	iHeaderLineCount ;
int 	i, n, rc ;
zMemoType	*pMemo ;

	/* Allocate memory */

	pMemo = (zMemoType*) XtCalloc ( sizeof(zMemoType), 1 ) ;
	if ( pMemo == NULL ) {
		xmemo_ErrorLog ( "xmemo_LoadMemo_FormatVersion_1", "XtCalloc", NULL ) ;
		return NULL ;
	}

	/* Line=0 : Number of header lines */

	i = 0 ;
	pLine[0] = NULL ;
	while ( i < 1  &&  *pBuf != 0 ) {
		pLine[i++] = pBuf ;
		while ( *pBuf != '\n'  &&  *pBuf != 0 ) pBuf ++ ;
		if ( *pBuf == '\n' ) {
			*pBuf = 0 ;
			pBuf ++ ;
		}
	}

	if ( ! xmemo_CheckNumericRange ( pLine[0], 11, 30 ) ) {
		xmemo_ErrorLog ( "xmemo_LoadMemo_FormatVersion_1", "Invalid", NULL ) ;
		xmemo_FreeMemo ( pMemo ) ;
		return NULL ;
	}

	iHeaderLineCount = atoi ( pLine[0] ) ;

	while ( i < iHeaderLineCount  &&  *pBuf != 0 ) {
		pLine[i++] = pBuf ;
		while ( *pBuf != '\n'  &&  *pBuf != 0 ) pBuf ++ ;
		if ( *pBuf == '\n' ) {
			*pBuf = 0 ;
			pBuf ++ ;
		}
	}

	if ( i != iHeaderLineCount ) {
		xmemo_ErrorLog ( "xmemo_LoadMemo_FormatVersion_1", "Invalid", NULL ) ;
		xmemo_FreeMemo ( pMemo ) ;
		return NULL ;
	}

	/* Line=1 : Transaction code */

	if ( strlen ( pLine[1] ) != 1 ) {
		xmemo_ErrorLog ( "xmemo_LoadMemo_FormatVersion_1", "Invalid", NULL ) ;
		xmemo_FreeMemo ( pMemo ) ;
		return NULL ;
	}

	pMemo->cTransaction = *pLine[1] ;

	/* Line=2 : x, y, w, h, iBorderWidth, iIconX, iIconY, iIconW, iIconH, iIconBorderWidth */

	n = xmemo_CsvToken ( pToken, 10, pLine[2] ) ;
	if ( n != 10
	  || ! xmemo_CheckNumeric ( pToken[0] ) 
	  || ! xmemo_CheckNumeric ( pToken[1] ) 
	  || ! xmemo_CheckNumeric ( pToken[2] ) 
	  || ! xmemo_CheckNumeric ( pToken[3] ) 
	  || ! xmemo_CheckNumeric ( pToken[4] ) 
	  || ! xmemo_CheckNumeric ( pToken[5] ) 
	  || ! xmemo_CheckNumeric ( pToken[6] ) 
	  || ! xmemo_CheckNumeric ( pToken[7] ) 
	  || ! xmemo_CheckNumeric ( pToken[8] ) 
	  || ! xmemo_CheckNumeric ( pToken[9] ) ) {
		xmemo_ErrorLog ( "xmemo_LoadMemo_FormatVersion_1", "Invalid", NULL ) ;
		xmemo_FreeMemo ( pMemo ) ;
		return NULL ;
	}

	pMemo->x = atoi ( pToken[0] ) ;
	pMemo->y = atoi ( pToken[1] ) ;
	pMemo->w = atoi ( pToken[2] ) ;
	pMemo->h = atoi ( pToken[3] ) ;
	pMemo->iBorderWidth = atoi ( pToken[4] ) ;
	pMemo->iIconX = atoi ( pToken[5] ) ;
	pMemo->iIconY = atoi ( pToken[6] ) ;
	pMemo->iIconW = atoi ( pToken[7] ) ;
	pMemo->iIconH = atoi ( pToken[8] ) ;
	pMemo->iIconBorderWidth = atoi ( pToken[9] ) ;

	/* Line=3 : bIcon, cIconPosition, iIconType */

	n = xmemo_CsvToken ( pToken, 3, pLine[3] ) ;
	if ( n != 3
	  || ! xmemo_CheckNumeric ( pToken[0] ) 
	  || ! xmemo_CheckNumeric ( pToken[1] ) 
	  || ! xmemo_CheckNumeric ( pToken[2] ) ) {
		xmemo_ErrorLog ( "xmemo_LoadMemo_FormatVersion_1", "Invalid", NULL ) ;
		xmemo_FreeMemo ( pMemo ) ;
		return NULL ;
	}

	pMemo->bIcon = ( atoi ( pToken[0] ) != 0 ) ;
	pMemo->cIconPosition = atoi ( pToken[1] ) ;
	pMemo->iIconType     = atoi ( pToken[2] ) ;

	/* Line=4 : iBorderColor, iBackColor, iFontColor */
	
	n = xmemo_CsvToken ( pToken, 9, pLine[4] ) ;
	if ( n != 9
	  || ! xmemo_CheckNumeric ( pToken[0] ) 
	  || ! xmemo_CheckNumeric ( pToken[1] ) 
	  || ! xmemo_CheckNumeric ( pToken[2] ) 
	  || ! xmemo_CheckNumeric ( pToken[3] ) 
	  || ! xmemo_CheckNumeric ( pToken[4] ) 
	  || ! xmemo_CheckNumeric ( pToken[5] ) 
	  || ! xmemo_CheckNumeric ( pToken[6] ) 
	  || ! xmemo_CheckNumeric ( pToken[7] ) 
	  || ! xmemo_CheckNumeric ( pToken[8] ) ) {
		xmemo_ErrorLog ( "xmemo_LoadMemo_FormatVersion_1", "Invalid", NULL ) ;
		xmemo_FreeMemo ( pMemo ) ;
		return NULL ;
	}

	pMemo->iBorderColor[0] = atoi ( pToken[0] ) ;
	pMemo->iBorderColor[1] = atoi ( pToken[1] ) ;
	pMemo->iBorderColor[2] = atoi ( pToken[2] ) ;
	pMemo->iBackColor[0] = atoi ( pToken[3] ) ;
	pMemo->iBackColor[1] = atoi ( pToken[4] ) ;
	pMemo->iBackColor[2] = atoi ( pToken[5] ) ;
	pMemo->iFontColor[0] = atoi ( pToken[6] ) ;
	pMemo->iFontColor[1] = atoi ( pToken[7] ) ;
	pMemo->iFontColor[2] = atoi ( pToken[8] ) ;

	/* Line=5 : pFontName */

	if ( strlen ( pLine[5] ) > 0 ) {
		pMemo->pFontName = XtMalloc ( strlen ( pLine[5] ) + 1 ) ;
		if ( pMemo->pFontName == NULL ) {
			xmemo_ErrorLog ( "xmemo_LoadMemo_FormatVersion_1", "XtMalloc", NULL ) ;
			xmemo_FreeMemo ( pMemo ) ;
			return NULL ;
		}
		strcpy ( pMemo->pFontName, pLine[5] ) ;
	}

	/* Line=6 : tCreate, iSerial, tExpire */

	n = xmemo_CsvToken ( pToken, 3, pLine[6] ) ;
	if ( n != 3
	  || ! xmemo_CheckNumeric ( pToken[0] ) 
	  || ! xmemo_CheckNumeric ( pToken[1] ) 
	  || ! xmemo_CheckNumeric ( pToken[2] ) ) {
		xmemo_ErrorLog ( "xmemo_LoadMemo_FormatVersion_1", "Invalid", NULL ) ;
		xmemo_FreeMemo ( pMemo ) ;
		return NULL ;
	}

	pMemo->tCreate = (time_t) atol ( pToken[0] ) ;
	pMemo->iSerial =          atoi ( pToken[1] ) ;
	pMemo->tExpire = (time_t) atol ( pToken[2] ) ;

	/* Line=7 : cProtocol, iAddr, iPort */

	n = xmemo_CsvToken ( pToken, 3, pLine[7] ) ;
	if ( n != 3
	  || strlen ( pToken[0] ) != 1
	  || ! xmemo_CheckNumericRange ( pToken[2], 0, 65535 ) ) {
		xmemo_ErrorLog ( "xmemo_LoadMemo_FormatVersion_1", "Invalid", NULL ) ;
		xmemo_FreeMemo ( pMemo ) ;
		return NULL ;
	}

	pMemo->cProtocol = *pToken[0] ;
	sscanf ( pToken[1], "%x", &(pMemo->iAddr) ) ;
	pMemo->iPort = atoi ( pToken[2] ) ;

	/* Line=8 : pHost */

	if ( strlen ( pLine[8] ) > 0 ) {
		pMemo->pHost = XtMalloc ( strlen ( pLine[8] ) + 1 ) ;
		if ( pMemo->pHost == NULL ) {
			xmemo_ErrorLog ( "xmemo_LoadMemo_FormatVersion_1", "XtMalloc", NULL ) ;
			xmemo_FreeMemo ( pMemo ) ;
			return NULL ;
		}
		strcpy ( pMemo->pHost, pLine[8] ) ;
	}

	/* Line=9 : pUser */

	if ( strlen ( pLine[9] ) > 0 ) {
		pMemo->pUser = XtMalloc ( strlen ( pLine[9] ) + 1 ) ;
		if ( pMemo->pUser == NULL ) {
			xmemo_ErrorLog ( "xmemo_LoadMemo_FormatVersion_1", "XtMalloc", NULL ) ;
			xmemo_FreeMemo ( pMemo ) ;
			return NULL ;
		}
		strcpy ( pMemo->pUser, pLine[9] ) ;
	}

	/* Line=10 : pNick */

	if ( strlen ( pLine[10] ) > 0 ) {
		pMemo->pNick = XtMalloc ( strlen ( pLine[10] ) + 1 ) ;
		if ( pMemo->pNick == NULL ) {
			xmemo_ErrorLog ( "xmemo_LoadMemo_FormatVersion_1", "XtMalloc", NULL ) ;
			xmemo_FreeMemo ( pMemo ) ;
			return NULL ;
		}
		strcpy ( pMemo->pNick, pLine[10] ) ;
	}

	/* Line=11 : pGroup */

	if ( strlen ( pLine[11] ) > 0 ) {
		pMemo->pGroup = XtMalloc ( strlen ( pLine[11] ) + 1 ) ;
		if ( pMemo->pGroup == NULL ) {
			xmemo_ErrorLog ( "xmemo_LoadMemo_FormatVersion_1", "XtMalloc", NULL ) ;
			xmemo_FreeMemo ( pMemo ) ;
			return NULL ;
		}
		strcpy ( pMemo->pGroup, pLine[11] ) ;
	}

	/* Lines until EOF : pText */

	if ( *pBuf != 0 ) {
		pMemo->pText = XtMalloc ( strlen ( pBuf ) + 1 ) ;
		if ( pMemo->pText == NULL ) {
			xmemo_ErrorLog ( "xmemo_LoadMemo_FormatVersion_1", "XtMalloc", NULL ) ;
			xmemo_FreeMemo ( pMemo ) ;
			return NULL ;
		}
		strcpy ( pMemo->pText, pBuf ) ;
	}

	return pMemo ;

}

/******************************************************************************/
/*  xmemo_LoadMemoAll                                                         */
/******************************************************************************/

int xmemo_LoadMemoAll ( )
{

char	sLog [ 100 ] ;
char	sBuf [ 500 + MAXLEN_HOSTNAME + MAXLEN_USERNAME + MAXLEN_NICKNAME + MAXLEN_GROUPNAME + MAXLEN_MESSAGE ] ;
char	fn [ MAXLEN_FILENAME ] ;
FILE	*fp ;
char	*pCh ;
int 	i, n, rc ;
char	*pToken [ 10 ] ;
int 	iLength, iFormat, iCypher ;
zMemoType	*pMemo, *pMemoTransaction ;
zChainType	*pChain ;
int 	bShrinkFiles ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_LoadMemoAll", NULL ) ;

	bShrinkFiles = FALSE ;

	for ( i = 1 ; i <= 2 ; i ++ ) {

		strcpy ( fn, sMemoBaseDir ) ;
		strcat ( fn, "/" ) ;
		strcat ( fn, sMemoDataDir ) ;
		strcat ( fn, "/" ) ;
		switch ( i ) {
		case 1 :
			strcat ( fn, sFilenameMemoArchives ) ;
			break ;
		case 2 :
			strcat ( fn, sFilenameMemoTransactions ) ;
			break ;
		}
		fp = fopen ( fn, "r" ) ;
		if ( fp == NULL ) continue ;

		if ( i == 2 ) {
			/* If transaction file exists, then merge the archives and */
			/* the transactions into archives after loading.           */
			bShrinkFiles = TRUE ;
		}

		for (;;) {

			if ( fgets ( sBuf, sizeof(sBuf), fp ) == NULL ) break ;
			pCh = strchr ( sBuf, '\n' ) ;
			if ( pCh != NULL ) *pCh = 0 ;

			/* Record length, Format version, Cypher type */
			n = xmemo_CsvToken ( pToken, 3, sBuf ) ;
			if ( n == 3
			  && xmemo_CheckNumericRange ( pToken[0], 0, sizeof(sBuf) ) 
			  && xmemo_CheckNumericRange ( pToken[1], 1, 1 )
			  && xmemo_CheckNumericRange ( pToken[2], 0, 0 ) ) {
				iLength = atoi ( pToken[0] ) ;
				iFormat = atoi ( pToken[1] ) ;
				iCypher = atoi ( pToken[2] ) ;
			} else {
				sprintf ( sLog, "File=[%s]", sFilenameMemoTransactions ) ;
				xmemo_ErrorLog ( "xmemo_LoadMemoAll", "Invalid", sLog ) ;
				break ;
			}

			/* Read one logical record */
			rc = fread ( sBuf, 1, iLength, fp ) ;
			if ( rc != iLength ) {
				sprintf ( sLog, "File=[%s]", sFilenameMemoTransactions ) ;
				xmemo_ErrorLog ( "xmemo_LoadMemoAll", "Invalid", sLog ) ;
			}
			sBuf[iLength] = 0 ;

			/* Build memo data */
			pMemoTransaction = xmemo_LoadMemo_FormatVersion_1 ( sBuf ) ;
			if ( pMemoTransaction == NULL ) {
				break ;
			}

			/* Remove old data */
			pChain = zMemoAnchor.pHead ;
			while ( pChain != NULL ) {

				pMemo = (zMemoType*) ( pChain->pData ) ;
				pChain = pChain->pNext ;

				if ( pMemo->tCreate == pMemoTransaction->tCreate
				  && pMemo->iSerial == pMemoTransaction->iSerial ) {

					/* Remove from chained list */
					xmemo_RemoveFromChain ( &zMemoAnchor, pMemo->pChain ) ;

					/* Free allocated memory */
					xmemo_FreeMemo ( pMemo ) ;

				}

			}

			if ( pMemoTransaction->cTransaction != XMEMO_TRANSACTION_DELETE ) {

				/* Insert into chained list */
				pMemoTransaction->pChain = xmemo_InsertIntoChain ( &zMemoAnchor, pMemoTransaction ) ;
				if ( pMemoTransaction->pChain == NULL ) {
					xmemo_ErrorLog ( "xmemo_LoadMemoAll", "xmemo_InsertIntoChain", NULL ) ;
					xmemo_FreeMemo ( pMemoTransaction ) ;
					break ;
				}

			}

		}

		fclose ( fp ) ;

	}

	if ( bShrinkFiles ) xmemo_SaveMemoAll ( ) ;

	return RC_SUCCESS ;

}

/******************************************************************************/
/*  xmemo_SaveMemo_FormatVersion_1                                            */
/******************************************************************************/

static int xmemo_SaveMemo_FormatVersion_1 ( FILE *fp, zMemoType *pMemo )
{

char	sLineBuf [ 10 ] [ 160 ] ;
int 	iLen, rc ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_SaveMemo", NULL ) ;

	/* Output length in byte */

	iLen = 0 ;

	/* Line=0 : Number of header lines */
	sprintf ( sLineBuf[0], "%d\n", 12 ) ;
	iLen += strlen ( sLineBuf[0] ) ;

	/* Line=1 : Transaction code */
	sprintf ( sLineBuf[1], "%c\n", pMemo->cTransaction ) ;
	iLen += strlen ( sLineBuf[1] ) ;

	/* Line=2 : x, y, w, h, iBorderWidth */
	sprintf ( sLineBuf[2], "%d,%d,%d,%d,%d,%d,%d,%d,%d,%d\n",
	          pMemo->x, pMemo->y, pMemo->w, pMemo->h, pMemo->iBorderWidth,
	          pMemo->iIconX, pMemo->iIconY, pMemo->iIconW, pMemo->iIconH, pMemo->iIconBorderWidth ) ;
	iLen += strlen ( sLineBuf[2] ) ;

	/* Line=3 : bIcon, cIconPosition, iIconType */
	sprintf ( sLineBuf[3], "%d,%d,%d\n",
	          (int) (pMemo->bIcon?1:0), (int) pMemo->cIconPosition, (int) pMemo->iIconType ) ;
	iLen += strlen ( sLineBuf[3] ) ;

	/* Line=4 : iBorderColor, iBackColor, iFontColor */
	sprintf ( sLineBuf[4], "%d,%d,%d,%d,%d,%d,%d,%d,%d\n",
	          pMemo->iBorderColor[0], pMemo->iBorderColor[1], pMemo->iBorderColor[2],
	          pMemo->iBackColor[0], pMemo->iBackColor[1], pMemo->iBackColor[2],
	          pMemo->iFontColor[0], pMemo->iFontColor[1], pMemo->iFontColor[2] ) ;
	iLen += strlen ( sLineBuf[4] ) ;

	/* Line=5 : pFontName */
	if ( pMemo->pFontName != NULL ) iLen += strlen ( pMemo->pFontName ) ;
	iLen += strlen ( "\n" ) ;

	/* Line=6 : tCreate, iSerial, tExpire */
	sprintf ( sLineBuf[6], "%ld,%d,%ld\n", pMemo->tCreate, pMemo->iSerial, pMemo->tExpire ) ;
	iLen += strlen ( sLineBuf[6] ) ;

	/* Line=7 : cProtocol, iAddr, iPort */
	sprintf ( sLineBuf[7], "%c,%08X,%d\n", pMemo->cProtocol,  pMemo->iAddr, pMemo->iPort ) ;
	iLen += strlen ( sLineBuf[7] ) ;

	/* Line=8 : pHost */
	if ( pMemo->pHost != NULL) iLen += strlen ( pMemo->pHost ) ;
	iLen += strlen ( "\n" ) ;

	/* Line=9 : pUser */
	if ( pMemo->pUser != NULL) iLen += strlen ( pMemo->pUser ) ;
	iLen += strlen ( "\n" ) ;

	/* Line=10 : pNick */
	if ( pMemo->pNick != NULL) iLen += strlen ( pMemo->pNick ) ;
	iLen += strlen ( "\n" ) ;

	/* Line=11 : pGroup */
	if ( pMemo->pGroup != NULL) iLen += strlen ( pMemo->pGroup ) ;
	iLen += strlen ( "\n" ) ;

	/* Line>12 : pText */
	iLen += strlen ( pMemo->pText ) ;

	/* Write */

	rc = fprintf ( fp, "%d,1,0\n", iLen ) ;

	/* Line=0 : Number of header lines */
	rc = fputs ( sLineBuf[0], fp ) ;

	/* Line=1 : Transaction code */
	rc = fputs ( sLineBuf[1], fp ) ;

	/* Line=2 : x, y, w, h, iBorderWidth */
	rc = fputs ( sLineBuf[2], fp ) ;

	/* Line=3 : bIcon, cIconPosition, iIconType */
	rc = fputs ( sLineBuf[3], fp ) ;

	/* Line=4 : iBorderColor, iBackColor, iFontColor */
	rc = fputs ( sLineBuf[4], fp ) ;

	/* Line=5 : pFontName */
	if ( pMemo->pFontName != NULL ) rc = fprintf ( fp, "%s\n", pMemo->pFontName ) ;
	else                            rc = fprintf ( fp, "\n" ) ;

	/* Line=6 : tCreate, iSerial, tExpire */
	rc = fputs ( sLineBuf[6], fp ) ;

	/* Line=7 : cProtocol, iAddr, iPort */
	rc = fputs ( sLineBuf[7], fp ) ;

	/* Line=8 : pHost */
	if ( pMemo->pHost != NULL) rc = fprintf ( fp, "%s\n", pMemo->pHost ) ;
	else                       rc = fprintf ( fp, "\n" ) ;

	/* Line=9 : pUser */
	if ( pMemo->pUser != NULL) rc = fprintf ( fp, "%s\n", pMemo->pUser ) ;
	else                       rc = fprintf ( fp, "\n" ) ;

	/* Line=10 : pNick */
	if ( pMemo->pNick != NULL) rc = fprintf ( fp, "%s\n", pMemo->pNick ) ;
	else                       rc = fprintf ( fp, "\n" ) ;

	/* Line=11 : pGroup */
	if ( pMemo->pGroup != NULL) rc = fprintf ( fp, "%s\n", pMemo->pGroup ) ;
	else                        rc = fprintf ( fp, "\n" ) ;

	/* Line>12 : pText */
	if ( pMemo->pText != NULL) rc = fputs ( pMemo->pText, fp ) ;

	return RC_SUCCESS ;

}

/******************************************************************************/
/*  xmemo_SaveMemoAll                                                         */
/******************************************************************************/

int xmemo_SaveMemoAll ( )
{

char	sLog [ 100 ] ;
char	fn [ MAXLEN_FILENAME ] ;
FILE	*fp ;
int 	rc ;
zChainType	*pChain ;
zMemoType	*pMemo ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_SaveMemoAll", NULL ) ;

	strcpy ( fn, sMemoBaseDir ) ;
	strcat ( fn, "/" ) ;
	strcat ( fn, sMemoDataDir ) ;
	strcat ( fn, "/" ) ;
	strcat ( fn, sFilenameMemoArchives ) ;
	fp = fopen ( fn, "w" ) ;
	if ( fp == NULL ) {
		xmemo_ErrorLog ( "xmemo_SaveMemoAll", "fopen", fn ) ;
		return RC_ERROR ;
	}

	for ( pChain = zMemoAnchor.pHead ; pChain != NULL ; pChain = pChain->pNext ) {

		xmemo_SaveMemo_FormatVersion_1 ( fp, (zMemoType*)(pChain->pData) ) ;

	}

	fclose ( fp ) ;

	strcpy ( fn, sMemoBaseDir ) ;
	strcat ( fn, "/" ) ;
	strcat ( fn, sMemoDataDir ) ;
	strcat ( fn, "/" ) ;
	strcat ( fn, sFilenameMemoTransactions ) ;
	remove ( fn ) ;

	return RC_SUCCESS ;

}

/******************************************************************************/
/*  xmemo_FreeMemoAll                                                         */
/******************************************************************************/

int xmemo_FreeMemoAll ( )
{

zMemoType		*pMemo ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_FreeMemoAll", NULL ) ;

	while ( zMemoAnchor.pHead != NULL ) {

		pMemo = (zMemoType*) ( zMemoAnchor.pHead->pData ) ;

		/* Remove from chained list */

		xmemo_RemoveFromChain ( &zMemoAnchor, zMemoAnchor.pHead ) ;

		/* Free allocated memory */

		xmemo_FreeMemo ( pMemo ) ;

	}

}

/******************************************************************************/
/*  xmemo_CreateMemo                                                          */
/******************************************************************************/

zMemoType* xmemo_CreateMemo ( )
{

int 	i ;
char	fn [ MAXLEN_FILENAME ] ;
zMemoType	*pMemo, *pMemoScan ;
zChainType	*pChain ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_CreateMemo", NULL ) ;

	/* Allocate memory */

	pMemo = (zMemoType*) XtCalloc ( sizeof(zMemoType), 1 ) ;
	if ( pMemo == NULL ) {
		xmemo_ErrorLog ( "xmemo_CreateMemo", "XtCalloc", NULL ) ;
		return NULL ;
	}

	/* Set default values */

	pMemo->w = iDefaultMemoWidth ;
	pMemo->h = iDefaultMemoHeight ;
	pMemo->iBorderWidth  = iDefaultMemoBorderWidth ;
	pMemo->cIconPosition = cDefaultMemoIconPosition ;
	pMemo->iIconType     = iDefaultMemoIconType ;
	time ( &(pMemo->tCreate) ) ;
	for ( pChain = zMemoAnchor.pHead ; pChain != NULL ; pChain = pChain->pNext ) {
		pMemoScan = (zMemoType*) pChain->pData ;
		if ( pMemo->tCreate == pMemoScan->tCreate
		  && pMemo->iSerial <= pMemoScan->iSerial ) {
			pMemo->iSerial = pMemoScan->iSerial + 1 ;
		}
	}
	for ( i = 0 ; i < 3 ; i ++ ) {
		pMemo->iBorderColor[i] = iDefaultMemoBorderColor[i] ;
		pMemo->iBackColor[i]   = iDefaultMemoBackColor[i] ;
		pMemo->iFontColor[i]   = iDefaultMemoFontColor[i] ;
	}
	pMemo->cProtocol = XMEMO_PROTOCOL_INVALID ;

	return pMemo ;

}

/******************************************************************************/
/*  xmemo_CopyMemo                                                            */
/******************************************************************************/

zMemoType* xmemo_CopyMemo ( zMemoType *pMemo )
{

int 	i ;
char	fn [ MAXLEN_FILENAME ] ;
zMemoType	*pCopyMemo, *pMemoScan ;
zChainType	*pChain ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_CopyMemo", NULL ) ;

	/* Allocate memory */

	pCopyMemo = (zMemoType*) XtCalloc ( sizeof(zMemoType), 1 ) ;
	if ( pCopyMemo == NULL ) {
		xmemo_ErrorLog ( "xmemo_CopyMemo", "XtCalloc", NULL ) ;
		return NULL ;
	}

	/* Set default values */

	pCopyMemo->x = pMemo->x ;
	pCopyMemo->y = pMemo->y ;
	pCopyMemo->w = pMemo->w ;
	pCopyMemo->h = pMemo->h ;
	pCopyMemo->iBorderWidth = pMemo->iBorderWidth ;

	time ( &(pCopyMemo->tCreate) ) ;
	for ( pChain = zMemoAnchor.pHead ; pChain != NULL ; pChain = pChain->pNext ) {
		pMemoScan = (zMemoType*) pChain->pData ;
		if ( pMemo->tCreate == pMemoScan->tCreate
		  && pMemo->iSerial <= pMemoScan->iSerial ) {
			pMemo->iSerial = pMemoScan->iSerial + 1 ;
		}
	}

	for ( i = 0 ; i < 3 ; i ++ ) {
		pCopyMemo->iBorderColor[i] = pMemo->iBorderColor[i] ;
		pCopyMemo->iBackColor[i]   = pMemo->iBackColor[i] ;
		pCopyMemo->iFontColor[i]   = pMemo->iFontColor[i] ;
	}

	pCopyMemo->cProtocol = pMemo->cProtocol ;

	if ( pMemo->pFontName != NULL ) {
		pCopyMemo->pFontName = XtMalloc ( strlen ( pMemo->pFontName ) + 1 ) ;
		if ( pCopyMemo->pFontName == NULL ) {
			xmemo_ErrorLog ( "xmemo_CopyMemo", "XtMalloc", NULL ) ;
			return NULL ;
		}
		strcpy ( pCopyMemo->pFontName, pMemo->pFontName ) ;
	}

	if ( pMemo->pText != NULL ) {
		pCopyMemo->pText = XtMalloc ( strlen ( pMemo->pText ) + 1 ) ;
		if ( pCopyMemo->pText == NULL ) {
			xmemo_ErrorLog ( "xmemo_CopyMemo", "XtMalloc", NULL ) ;
			return NULL ;
		}
		strcpy ( pCopyMemo->pText, pMemo->pText ) ;
	}

	return pCopyMemo ;

}

/******************************************************************************/
/*  xmemo_InsertMemo                                                          */
/******************************************************************************/

int xmemo_InsertMemo ( zMemoType *pMemo )
{

char	fn [ MAXLEN_FILENAME ] ;
FILE	*fp ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_InsertMemo", NULL ) ;

	/* Insert into chained list */

	pMemo->pChain = xmemo_InsertIntoChain ( &zMemoAnchor, pMemo ) ;
	if ( pMemo->pChain == NULL ) {
		xmemo_ErrorLog ( "xmemo_InsertMemo", "xmemo_InsertIntoChain", NULL ) ;
		xmemo_FreeMemo ( pMemo ) ;
		return RC_ERROR ;
	}

	/* Create widget */
	
	xmemo_CreateMemoWidget ( pMemo ) ;

	/* Insert transaction */

	pMemo->cTransaction = XMEMO_TRANSACTION_INSERT ;

	strcpy ( fn, sMemoBaseDir ) ;
	strcat ( fn, "/" ) ;
	strcat ( fn, sMemoDataDir ) ;
	strcat ( fn, "/" ) ;
	strcat ( fn, sFilenameMemoTransactions ) ;
	fp = fopen ( fn, "a" ) ;
	if ( fp != NULL ) {
		xmemo_SaveMemo_FormatVersion_1 ( fp, pMemo ) ;
		fclose ( fp ) ;
	} else {
		xmemo_ErrorLog ( "xmemo_SaveMemo", "fopen", fn ) ;
		return RC_ERROR ;
	}

	return RC_SUCCESS ;

}

/******************************************************************************/
/*  xmemo_DeleteMemo                                                          */
/******************************************************************************/

int xmemo_DeleteMemo ( zMemoType *pMemo )
{

char	fn [ MAXLEN_FILENAME ] ;
FILE	*fp ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_DeleteMemo", NULL ) ;

	/* Destroy Widget */
	
	xmemo_DestroyMemoWidget ( pMemo ) ;

	/* Remove from chained list */

	if ( pMemo->pChain != NULL ) {
		xmemo_RemoveFromChain ( &zMemoAnchor, pMemo->pChain ) ;
	}

	/* Delete transaction */

	pMemo->cTransaction = XMEMO_TRANSACTION_DELETE ;

	strcpy ( fn, sMemoBaseDir ) ;
	strcat ( fn, "/" ) ;
	strcat ( fn, sMemoDataDir ) ;
	strcat ( fn, "/" ) ;
	strcat ( fn, sFilenameMemoTransactions ) ;
	fp = fopen ( fn, "a" ) ;
	if ( fp != NULL ) {
		xmemo_SaveMemo_FormatVersion_1 ( fp, pMemo ) ;
		fclose ( fp ) ;
	} else {
		xmemo_ErrorLog ( "xmemo_SaveMemo", "fopen", fn ) ;
		return RC_ERROR ;
	}

	/* Free allocated memory */

	xmemo_FreeMemo ( pMemo ) ;

	return RC_SUCCESS ;

}

/******************************************************************************/
/*  xmemo_UpdateMemo                                                          */
/******************************************************************************/

int xmemo_UpdateMemo ( zMemoType *pMemo )
{

char	fn [ MAXLEN_FILENAME ] ;
FILE	*fp ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_UpdateMemo", NULL ) ;

	/* Update transaction */

	pMemo->cTransaction = XMEMO_TRANSACTION_UPDATE ;

	strcpy ( fn, sMemoBaseDir ) ;
	strcat ( fn, "/" ) ;
	strcat ( fn, sMemoDataDir ) ;
	strcat ( fn, "/" ) ;
	strcat ( fn, sFilenameMemoTransactions ) ;
	fp = fopen ( fn, "a" ) ;
	if ( fp != NULL ) {
		xmemo_SaveMemo_FormatVersion_1 ( fp, pMemo ) ;
		fclose ( fp ) ;
	} else {
		xmemo_ErrorLog ( "xmemo_SaveMemo", "fopen", fn ) ;
		return RC_ERROR ;
	}

	return RC_SUCCESS ;

}
