/******************************************************************************/
/*                                                                            */
/*  <<< Copyright >>>                                                         */
/*                                                                            */
/*  The "xmemo" software is copyright 1999 by Takao Abe.                      */
/*  All rights reserved.                                                      */
/*                                                                            */
/*  The "xmemo" software refers to the source code, the binary code,          */
/*  the documents, and all derivatives based on this software under           */
/*  the copyright law, that is a software containing a program or a           */
/*  portion of a program, regardless of modifications and/or translations.    */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  <<< Licence and redistribution >>>                                        */
/*                                                                            */
/*  The source code and the binary code are free software.                    */
/*  The use and redistribution of source code and binary code are permitted   */
/*  provided that the following conditions are met:                           */
/*  You may use the software for private use and commercial use,              */
/*  and may modify your copy or copies of the software or any portion of it.  */
/*                                                                            */
/*  THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESSED OR IMPLIED            */
/*  WARRANTIES OF ANY KIND, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED        */
/*  WARRANTIES OF MERCHANTABLILITY AND FITNESS FOR A PARTICULAR PURPOSE.      */
/*  IN NO EVENT SHALL THE AUTHOR TAKAO ABE BE LIABLE FOR ANY DIRECT,          */
/*  INDIRECT, GENERAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES           */
/*  ( INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR       */
/*  SERVICES; LOSS OF USE, DATA OR PROFITS; OR BUSINESS INTERRUPTION )        */
/*  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,       */
/*  STRICT LIABILITY, OR TORT ( INCLUDING NEGLIGENCE OR OTHERWISE )           */
/*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED       */
/*  OF THE POSSIBILITY OF SUCH DAMAGE.                                        */
/*                                                                            */
/*  Redistribution and use in source form or binary form, with or without     */
/*  modification, the whole program or a portion of it, are permitted         */
/*  provided that one retains the entire copyright notice properly, and       */
/*  displays or retains this license in the documentation and/or other        */
/*  materials provided with the distribution.                                 */
/*                                                                            */
/*  Redistribution and use in source form or binary form, with modification,  */
/*  the whole program or a portion of it, are permitted provided that one     */
/*  declares that it is modified software and not original, and displays or   */
/*  describes any modifications in the documentation and/or other materials   */
/*  provided with the distribution.                                           */
/*                                                                            */
/*  This software and the name of the author must not be used to endorse or   */
/*  promote products derived from this software without prior written         */
/*  permission.                                                               */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  EMail   abetakao@bea.hi-ho.ne.jp                                          */
/*          HCA00763@nifty.ne.jp                                              */
/*                                                                            */
/*  URL     http://www.bea.hi-ho.ne.jp/abetakao/                              */
/*                                                                            */
/******************************************************************************/


#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<time.h>


#include	<X11/Intrinsic.h>


#include	"xmemo.h"
#include	"xmemodata.h"
#include	"xmemofunc.h"


static	time_t	tTimer = 0 ;

/******************************************************************************/
/*  Local constant definition                                                 */
/******************************************************************************/

#define		XMEMO_TIMEOUT_TCP_CONNECT	10
#define		XMEMO_TIMEOUT_SEND      	5
#define		XMEMO_TIMEOUT_RECV      	5

#define		XMEMO_RETRYOUT          	3

/******************************************************************************/
/*  xmemo_Timer                                                               */
/******************************************************************************/

static void xmemo_Timer ( caddr_t pClientData, XtIntervalId* pId )
{

zChainType	*pChain ;
zMemberType	*pPeer ;
zSocketType	*pSocket ;
time_t	tNow, tNext ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_Timer", NULL ) ;

	time ( &tNow ) ;

	tTimer = 0 ;

	pChain = zSendPeerAnchor.pHead ;
	while ( pChain != NULL ) {

		pPeer = (zMemberType*) ( pChain->pData ) ;
		pChain = pChain->pNext ;
		if ( pPeer->tTimeout == 0  ||  pPeer->tTimeout > tNow ) continue ;

		switch ( pPeer->cProtocol ) {
		case XMEMO_PROTOCOL_HUSEN :
			xmemo_TimeoutUdp_Husen ( pPeer ) ;
			break ;
		case XMEMO_PROTOCOL_IPMSG :
			xmemo_TimeoutUdp_Ipmsg ( pPeer ) ;
			break ;
		case XMEMO_PROTOCOL_OBOEGAKI :
			xmemo_TimeoutConnectedTcp_Oboegaki ( pPeer ) ;
			break ;
		}

	}

	pChain = zSocketAnchor.pHead ;
	while ( pChain != NULL ) {

		pSocket = (zSocketType*) ( pChain->pData ) ;
		pChain = pChain->pNext ;
		if ( pSocket->tTimeout == 0  ||  pSocket->tTimeout > tNow ) continue ;

		switch ( pSocket->cProtocol ) {
		case XMEMO_PROTOCOL_HUSEN :
			break ;
		case XMEMO_PROTOCOL_IPMSG :
			break ;
		case XMEMO_PROTOCOL_OBOEGAKI :
			xmemo_TimeoutAcceptedTcp_Oboegaki ( pSocket ) ;
			break ;
		}

	}

	/* Next timeout */

	tNext = 0 ;

	pChain = zSendPeerAnchor.pHead ;
	while ( pChain != NULL ) {

		pPeer = (zMemberType*) ( pChain->pData ) ;
		pChain = pChain->pNext ;
		if ( pPeer->tTimeout != 0  &&  ( pPeer->tTimeout < tNext  ||  tNext == 0 ) ) tNext = pPeer->tTimeout ;

	}

	pChain = zSocketAnchor.pHead ;
	while ( pChain != NULL ) {

		pSocket = (zSocketType*) ( pChain->pData ) ;
		pChain = pChain->pNext ;
		if ( pSocket->tTimeout != 0  &&  ( pSocket->tTimeout < tNext  ||  tNext == 0 ) ) tNext = pSocket->tTimeout ;

	}

	if ( tNext != 0 ) xmemo_StartTimer ( tNext ) ;

	return ;

}

/******************************************************************************/
/*  xmemo_StartTimer                                                          */
/******************************************************************************/

void xmemo_StartTimer ( time_t tTimeout )
{

time_t	t ;
time_t	tInterval ;

	if ( tTimeout >= tTimer  &&  tTimer != 0 ) return ;

	if ( bMemoTraceLog ) {
		char	sLog [ 32 ] ;
		struct	tm	*pYMDHMS ;
		pYMDHMS = localtime ( &tTimeout ) ;
		sprintf ( sLog, "%04d/%02d/%02d  %02d:%02d:%02d",
		          pYMDHMS->tm_year+1900, pYMDHMS->tm_mon+1, pYMDHMS->tm_mday,
		          pYMDHMS->tm_hour, pYMDHMS->tm_min, pYMDHMS->tm_sec ) ;
		xmemo_TraceLog ( "xmemo_StartTimer", sLog ) ;	
	}

	time ( &t ) ;
	tInterval = ( tTimeout - t ) * 1000 ;
	if ( tInterval <= 0 ) tInterval = 1000 ;
	tTimer = tTimeout ;

	XtAppAddTimeOut ( xApp, tInterval, (XtTimerCallbackProc) xmemo_Timer, NULL ) ;

}
