/******************************************************************************/
/*                                                                            */
/*  <<< Copyright >>>                                                         */
/*                                                                            */
/*  The "xmemo" software is copyright 1999 by Takao Abe.                      */
/*  All rights reserved.                                                      */
/*                                                                            */
/*  The "xmemo" software refers to the source code, the binary code,          */
/*  the documents, and all derivatives based on this software under           */
/*  the copyright law, that is a software containing a program or a           */
/*  portion of a program, regardless of modifications and/or translations.    */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  <<< Licence and redistribution >>>                                        */
/*                                                                            */
/*  The source code and the binary code are free software.                    */
/*  The use and redistribution of source code and binary code are permitted   */
/*  provided that the following conditions are met:                           */
/*  You may use the software for private use and commercial use,              */
/*  and may modify your copy or copies of the software or any portion of it.  */
/*                                                                            */
/*  THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESSED OR IMPLIED            */
/*  WARRANTIES OF ANY KIND, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED        */
/*  WARRANTIES OF MERCHANTABLILITY AND FITNESS FOR A PARTICULAR PURPOSE.      */
/*  IN NO EVENT SHALL THE AUTHOR TAKAO ABE BE LIABLE FOR ANY DIRECT,          */
/*  INDIRECT, GENERAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES           */
/*  ( INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR       */
/*  SERVICES; LOSS OF USE, DATA OR PROFITS; OR BUSINESS INTERRUPTION )        */
/*  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,       */
/*  STRICT LIABILITY, OR TORT ( INCLUDING NEGLIGENCE OR OTHERWISE )           */
/*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED       */
/*  OF THE POSSIBILITY OF SUCH DAMAGE.                                        */
/*                                                                            */
/*  Redistribution and use in source form or binary form, with or without     */
/*  modification, the whole program or a portion of it, are permitted         */
/*  provided that one retains the entire copyright notice properly, and       */
/*  displays or retains this license in the documentation and/or other        */
/*  materials provided with the distribution.                                 */
/*                                                                            */
/*  Redistribution and use in source form or binary form, with modification,  */
/*  the whole program or a portion of it, are permitted provided that one     */
/*  declares that it is modified software and not original, and displays or   */
/*  describes any modifications in the documentation and/or other materials   */
/*  provided with the distribution.                                           */
/*                                                                            */
/*  This software and the name of the author must not be used to endorse or   */
/*  promote products derived from this software without prior written         */
/*  permission.                                                               */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  EMail   abetakao@bea.hi-ho.ne.jp                                          */
/*          HCA00763@nifty.ne.jp                                              */
/*                                                                            */
/*  URL     http://www.bea.hi-ho.ne.jp/abetakao/                              */
/*                                                                            */
/******************************************************************************/


#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<time.h>


#include	<X11/Xlib.h>
#include	<X11/Xutil.h>
#include	<X11/Intrinsic.h>
#include	<X11/StringDefs.h>
#include	<X11/Shell.h>
#include	<X11/cursorfont.h>

#include	<X11/Xaw/AsciiText.h>
#include	<X11/Xaw/Form.h>
#include	<X11/Xaw/MenuButton.h>
#include	<X11/Xaw/Simple.h>
#include	<X11/Xaw/SimpleMenu.h>
#include	<X11/Xaw/SmeBSB.h>
#include	<X11/Xaw/SmeLine.h>


#include	"xmemo.h"
#include	"xmemodata.h"
#include	"xmemofunc.h"


/******************************************************************************/
/*  Local variables                                                           */
/******************************************************************************/


static	Widget	wMemoPopupMenu ;

static	Cursor	xBorderCursorRight ;
static	Cursor	xBorderCursorBottom ;
static	Cursor	xBorderCursorCorner ;


/******************************************************************************/
/*  MemoEvent_ButtonPress                                                     */
/******************************************************************************/

static void MemoEvent_ButtonPress ( Widget w, caddr_t pClientData, XEvent *pEvent )
{

zMemoType		*pMemo ;
zChainType		*pChain ;

Widget  wi ;
int 	n ;
Arg 	xArg [ 10 ] ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "MemoEvent_ButtonPress", NULL ) ;

	for ( pChain = zMemoAnchor.pHead ; pChain != NULL ; pChain = pChain->pNext ) {

		pMemo = (zMemoType*) ( pChain->pData ) ;
		if ( pMemo->wText == w ) {

			XRaiseWindow ( XtDisplay(pMemo->wMemo), XtWindow(pMemo->wMemo) ) ;
			pMemo->iMouseRootX = pEvent->xmotion.x_root ;
			pMemo->iMouseRootY = pEvent->xmotion.y_root ;
			pMemo->bMouseDrag = False ;

			if ( pEvent->xbutton.button == Button3 ) {
				pMemoActive = pMemo ;
				XtMoveWidget ( wMemoPopupMenu, pMemo->iMouseRootX, pMemo->iMouseRootY ) ;
				XtPopup ( wMemoPopupMenu, XtGrabExclusive ) ;
			}

			break ;
		}

	}

}

/******************************************************************************/
/*  MemoEvent_ButtonRelease                                                   */
/******************************************************************************/

static void MemoEvent_ButtonRelease ( Widget w, caddr_t pClientData, XEvent *pEvent )
{

zMemoType		*pMemo ;
zChainType		*pChain ;

Widget  wi ;
int 	n ;
Arg 	xArg [ 10 ] ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "MemoEvent_ButtonRelease", NULL ) ;

	for ( pChain = zMemoAnchor.pHead ; pChain != NULL ; pChain = pChain->pNext ) {

		pMemo = (zMemoType*) ( pChain->pData ) ;
		if ( pMemo->wText == w ) {

			if ( pEvent->xbutton.button == Button1  &&  pMemo->bMouseDrag ) {
				pMemo->x += pEvent->xmotion.x_root - pMemo->iMouseRootX ;
				pMemo->y += pEvent->xmotion.y_root - pMemo->iMouseRootY ;
				XtMoveWidget ( pMemo->wMemo, pMemo->x, pMemo->y ) ;
				pMemo->iMouseRootX = pEvent->xmotion.x_root ;
				pMemo->iMouseRootY = pEvent->xmotion.y_root ;
				pMemo->bMouseDrag = False ;
				xmemo_UpdateMemo ( pMemo ) ;
			}

			break ;
		}

	}

}

/******************************************************************************/
/*  MemoEvent_Drag                                                            */
/******************************************************************************/

static void MemoEvent_Drag ( Widget w, caddr_t pClientData, XEvent *pEvent )
{

zMemoType		*pMemo ;
zChainType		*pChain ;

	if ( ( pEvent->xmotion.state & ( Button1Mask | Button2Mask ) ) == 0 ) return ;

#if 0
	if ( bMemoTraceLog ) xmemo_TraceLog ( "MemoEvent_Drag", NULL ) ;
#endif

	for ( pChain = zMemoAnchor.pHead ; pChain != NULL ; pChain = pChain->pNext ) {

		pMemo = (zMemoType*) ( pChain->pData ) ;
		if ( pMemo->wText == w ) {
			if ( ! pMemo->bMouseDrag ) {
				if ( abs( pEvent->xmotion.x_root - pMemo->iMouseRootX ) >= 2
				  || abs( pEvent->xmotion.y_root - pMemo->iMouseRootY ) >= 2 ) {
					pMemo->bMouseDrag = True ;
				}
			}
			if ( pMemo->bMouseDrag ) {
				pMemo->x += pEvent->xmotion.x_root - pMemo->iMouseRootX ;
				pMemo->y += pEvent->xmotion.y_root - pMemo->iMouseRootY ;
				XtMoveWidget ( pMemo->wMemo, pMemo->x, pMemo->y ) ;
				pMemo->iMouseRootX = pEvent->xmotion.x_root ;
				pMemo->iMouseRootY = pEvent->xmotion.y_root ;
			}
			break ;
		}

	}

}

/******************************************************************************/
/*  BorderEvent_ButtonPress                                                   */
/******************************************************************************/

static void BorderEvent_ButtonPress ( Widget w, caddr_t pClientData, XEvent *pEvent )
{

zMemoType		*pMemo ;
zChainType		*pChain ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "BorderEvent_ButtonPress", NULL ) ;

	for ( pChain = zMemoAnchor.pHead ; pChain != NULL ; pChain = pChain->pNext ) {

		pMemo = (zMemoType*) ( pChain->pData ) ;
		if ( pMemo->wBorderBottom == w || pMemo->wBorderRight == w || pMemo->wBorderCorner == w ) {

			XRaiseWindow ( XtDisplay(pMemo->wMemo), XtWindow(pMemo->wMemo) ) ;
			pMemo->iMouseRootX = pEvent->xmotion.x_root ;
			pMemo->iMouseRootY = pEvent->xmotion.y_root ;
			pMemo->bMouseDrag = False ;
			break ;

		}

	}

}

/******************************************************************************/
/*  BorderEvent_ButtonRelease                                                 */
/******************************************************************************/

static void BorderEvent_ButtonRelease ( Widget w, caddr_t pClientData, XEvent *pEvent )
{

zMemoType		*pMemo ;
zChainType		*pChain ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "BorderEvent_ButtonRelease", NULL ) ;

	for ( pChain = zMemoAnchor.pHead ; pChain != NULL ; pChain = pChain->pNext ) {

		pMemo = (zMemoType*) ( pChain->pData ) ;
		if ( pMemo->wBorderBottom == w ) {

			if ( pEvent->xbutton.button == Button1  &&  pMemo->bMouseDrag ) {
				pMemo->h += pEvent->xmotion.y_root - pMemo->iMouseRootY ;
				XtResizeWidget ( pMemo->wMemo, pMemo->w, pMemo->h, pMemo->iBorderWidth ) ;
				pMemo->iMouseRootX = pEvent->xmotion.x_root ;
				pMemo->iMouseRootY = pEvent->xmotion.y_root ;
				pMemo->bMouseDrag = False ;
				xmemo_UpdateMemo ( pMemo ) ;
			}
			break ;

		} else if ( pMemo->wBorderRight == w ) {

			if ( pEvent->xbutton.button == Button1  &&  pMemo->bMouseDrag ) {
				pMemo->w += pEvent->xmotion.x_root - pMemo->iMouseRootX ;
				XtResizeWidget ( pMemo->wMemo, pMemo->w, pMemo->h, pMemo->iBorderWidth ) ;
				pMemo->iMouseRootX = pEvent->xmotion.x_root ;
				pMemo->iMouseRootY = pEvent->xmotion.y_root ;
				pMemo->bMouseDrag = False ;
				xmemo_UpdateMemo ( pMemo ) ;
			}
			break ;

		} else if ( pMemo->wBorderCorner == w ) {

			if ( pEvent->xbutton.button == Button1  &&  pMemo->bMouseDrag ) {
				pMemo->w += pEvent->xmotion.x_root - pMemo->iMouseRootX ;
				pMemo->h += pEvent->xmotion.y_root - pMemo->iMouseRootY ;
				XtResizeWidget ( pMemo->wMemo, pMemo->w, pMemo->h, pMemo->iBorderWidth ) ;
				pMemo->iMouseRootX = pEvent->xmotion.x_root ;
				pMemo->iMouseRootY = pEvent->xmotion.y_root ;
				pMemo->bMouseDrag = False ;
				xmemo_UpdateMemo ( pMemo ) ;
			}
			break ;

		}

	}

}

/******************************************************************************/
/*  BorderEvent_Drag                                                          */
/******************************************************************************/

static void BorderEvent_Drag ( Widget w, caddr_t pClientData, XEvent *pEvent )
{

zMemoType		*pMemo ;
zChainType		*pChain ;

	if ( ( pEvent->xmotion.state & ( Button1Mask | Button2Mask | Button3Mask ) ) == 0 ) return ;

#if 0
	if ( bMemoTraceLog ) xmemo_TraceLog ( "BorderEvent_Drag", NULL ) ;
#endif

	for ( pChain = zMemoAnchor.pHead ; pChain != NULL ; pChain = pChain->pNext ) {

		pMemo = (zMemoType*) ( pChain->pData ) ;
		if ( pMemo->wBorderBottom == w ) {

			if ( ! pMemo->bMouseDrag ) {
				if ( abs( pEvent->xmotion.y_root - pMemo->iMouseRootY ) >= 2 ) {
					pMemo->bMouseDrag = True ;
				}
			}
			if ( pMemo->bMouseDrag ) {
				pMemo->h += pEvent->xmotion.y_root - pMemo->iMouseRootY ;
				XtResizeWidget ( pMemo->wMemo, pMemo->w, pMemo->h, pMemo->iBorderWidth ) ;
				pMemo->iMouseRootX = pEvent->xmotion.x_root ;
				pMemo->iMouseRootY = pEvent->xmotion.y_root ;
			}
			break ;

		} else if ( pMemo->wBorderRight == w ) {

			if ( ! pMemo->bMouseDrag ) {
				if ( abs( pEvent->xmotion.x_root - pMemo->iMouseRootX ) >= 2 ) {
					pMemo->bMouseDrag = True ;
				}
			}
			if ( pMemo->bMouseDrag ) {
				pMemo->w += pEvent->xmotion.x_root - pMemo->iMouseRootX ;
				XtResizeWidget ( pMemo->wMemo, pMemo->w, pMemo->h, pMemo->iBorderWidth ) ;
				pMemo->iMouseRootX = pEvent->xmotion.x_root ;
				pMemo->iMouseRootY = pEvent->xmotion.y_root ;
			}
			break ;

		} else if ( pMemo->wBorderCorner == w ) {

			if ( ! pMemo->bMouseDrag ) {
				if ( abs( pEvent->xmotion.x_root - pMemo->iMouseRootX ) >= 2
				  || abs( pEvent->xmotion.y_root - pMemo->iMouseRootY ) >= 2 ) {
					pMemo->bMouseDrag = True ;
				}
			}
			if ( pMemo->bMouseDrag ) {
				pMemo->w += pEvent->xmotion.x_root - pMemo->iMouseRootX ;
				pMemo->h += pEvent->xmotion.y_root - pMemo->iMouseRootY ;
				XtResizeWidget ( pMemo->wMemo, pMemo->w, pMemo->h, pMemo->iBorderWidth ) ;
				pMemo->iMouseRootX = pEvent->xmotion.x_root ;
				pMemo->iMouseRootY = pEvent->xmotion.y_root ;
			}
			break ;

		}

	}

}

/******************************************************************************/
/*  PopupEdit_Click                                                           */
/******************************************************************************/

static void PopupEdit_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

	if ( bMemoTraceLog ) xmemo_TraceLog ( "PopupEdit_Click", NULL ) ;

	xmemo_PopupDialog_Edit ( ) ;

}

static XtCallbackRec pPopupEdit_CallbackRec [ ] =
{
	{ (XtCallbackProc) PopupEdit_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  PopupAttribute_Click                                                      */
/******************************************************************************/

static void PopupAttribute_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

	if ( bMemoTraceLog ) xmemo_TraceLog ( "PopupAttribute_Click", NULL ) ;

	xmemo_PopupDialog_MemoAttribute ( pMemoActive ) ;

}

static XtCallbackRec pPopupAttribute_CallbackRec [ ] =
{
	{ (XtCallbackProc) PopupAttribute_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  PopupSend_Click                                                           */
/******************************************************************************/

static void PopupSend_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

	if ( bMemoTraceLog ) xmemo_TraceLog ( "PopupSend_Click", NULL ) ;

	xmemo_PopupDialog_Send ( ) ;

}

static XtCallbackRec pPopupSend_CallbackRec [ ] =
{
	{ (XtCallbackProc) PopupSend_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  PopupDelete_Click                                                         */
/******************************************************************************/

static void PopupDelete_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

	if ( bMemoTraceLog ) xmemo_TraceLog ( "PopupDelete_Click", NULL ) ;

	xmemo_DeleteMemo ( pMemoActive ) ;

}

static XtCallbackRec pPopupDelete_CallbackRec [ ] =
{
	{ (XtCallbackProc) PopupDelete_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  xmemo_InitWindow_Memo                                                     */
/******************************************************************************/

int xmemo_InitWindow_Memo ( )
{

int 	i ;
Arg	xArg [ 4 ] ;
int 	n ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_InitWindow_Memo", NULL ) ;

	/* Popup menu */

	wMemoPopupMenu = XtCreatePopupShell ( "xmemoPopupMenu0", simpleMenuWidgetClass, wTop, NULL, 0 ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel, "Edit" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback, pPopupEdit_CallbackRec ) ; n ++ ;
	XtCreateManagedWidget ( "cmd01", smeBSBObjectClass, wMemoPopupMenu, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel, "Attributes" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback, pPopupAttribute_CallbackRec ) ; n ++ ;
	XtCreateManagedWidget ( "cmd01", smeBSBObjectClass, wMemoPopupMenu, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel, "Send" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback, pPopupSend_CallbackRec ) ; n ++ ;
	XtCreateManagedWidget ( "cmd01", smeBSBObjectClass, wMemoPopupMenu, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel, "Delete" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback, pPopupDelete_CallbackRec ) ; n ++ ;
	XtCreateManagedWidget ( "cmd02", smeBSBObjectClass, wMemoPopupMenu, xArg, n ) ;

	XtCreateManagedWidget ( "cmdLine1", smeLineObjectClass, wMemoPopupMenu, NULL, 0 ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel, "Cancel" ) ; n ++ ;
	XtCreateManagedWidget ( "cmd02", smeBSBObjectClass, wMemoPopupMenu, xArg, n ) ;

	/* Border cursor */

	xBorderCursorRight  = XCreateFontCursor ( XtDisplay(wTop), XC_right_side ) ;
	xBorderCursorBottom = XCreateFontCursor ( XtDisplay(wTop), XC_bottom_side ) ;
	xBorderCursorCorner = XCreateFontCursor ( XtDisplay(wTop), XC_bottom_right_corner ) ;

	return RC_SUCCESS ;

}

/******************************************************************************/
/*  xmemo_SetMemoWidgetAttribute                                              */
/******************************************************************************/

int xmemo_SetMemoWidgetAttribute ( zMemoType *pMemo )
{

int 	n ;
Arg 	xArg [ 16 ] ;
Colormap	colormap ;
XColor  	backcolor, fontcolor, bordercolor ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_SetMemoWidgetAttribute", NULL ) ;

	XtResizeWidget ( pMemo->wMemo, pMemo->w, pMemo->h, pMemo->iBorderWidth ) ;

	colormap = XDefaultColormap ( XtDisplay(wTop), 0 ) ;
	bordercolor.red   = (unsigned short) pMemo->iBorderColor[0] ;
	bordercolor.green = (unsigned short) pMemo->iBorderColor[1] ;
	bordercolor.blue  = (unsigned short) pMemo->iBorderColor[2] ;
	XAllocColor ( XtDisplay(wTop), colormap, &bordercolor ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNborderColor, bordercolor.pixel ) ; n ++ ;
	XtSetValues ( pMemo->wMemo, xArg, n ) ;

	colormap = XDefaultColormap ( XtDisplay(wTop), 0 ) ;
	backcolor.red   = (unsigned short) pMemo->iBackColor[0] ;
	backcolor.green = (unsigned short) pMemo->iBackColor[1] ;
	backcolor.blue  = (unsigned short) pMemo->iBackColor[2] ;
	XAllocColor ( XtDisplay(wTop), colormap, &backcolor ) ;

	colormap = XDefaultColormap ( XtDisplay(wTop), 0 ) ;
	fontcolor.red   = (unsigned short) pMemo->iFontColor[0] ;
	fontcolor.green = (unsigned short) pMemo->iFontColor[1] ;
	fontcolor.blue  = (unsigned short) pMemo->iFontColor[2] ;
	XAllocColor ( XtDisplay(wTop), colormap, &fontcolor ) ;

	if ( pMemo->pFontStruct != NULL ) {
		XFreeFont ( XtDisplay(wTop), pMemo->pFontStruct ) ;
		pMemo->pFontStruct = NULL ;
	}
	if ( pMemo->pFontName != NULL ) {
		pMemo->pFontStruct = XLoadQueryFont ( XtDisplay(wTop), pMemo->pFontName ) ;
		if ( pMemo->pFontStruct == NULL  &&  bMemoTraceLog ) {
			xmemo_TraceLog ( "xmemo_SetMemoWidgetAttribute", "XLoadQueryFont == NULL" ) ;
		}
	} else {
		pMemo->pFontStruct = NULL ;
	}

	n = 0 ;
	XtSetArg ( xArg[n], XtNbackground, backcolor.pixel ) ; n ++ ;
	XtSetArg ( xArg[n], XtNforeground, fontcolor.pixel ) ; n ++ ;
	if ( pMemo->pFontStruct != NULL ) {
		XtSetArg ( xArg[n], XtNfont, pMemo->pFontStruct ) ; n ++ ;
	}
	XtSetValues ( pMemo->wText, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNbackground, backcolor.pixel ) ; n ++ ;
	XtSetValues ( pMemo->wBorderBottom, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNbackground, backcolor.pixel ) ; n ++ ;
	XtSetValues ( pMemo->wBorderRight, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNbackground, backcolor.pixel ) ; n ++ ;
	XtSetValues ( pMemo->wBorderCorner, xArg, n ) ;

}

/******************************************************************************/
/*  xmemo_CreateMemoWidget                                                    */
/******************************************************************************/

int xmemo_CreateMemoWidget ( zMemoType *pMemo )
{

int 	n ;
Arg 	xArg [ 20 ] ;
Colormap	colormap ;
XColor  	backcolor, fontcolor, bordercolor ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_CreateMemoWidget", NULL ) ;

	/* Memo Shell */

	colormap = XDefaultColormap ( XtDisplay(wTop), 0 ) ;
	bordercolor.red   = (unsigned short) pMemo->iBorderColor[0] ;
	bordercolor.green = (unsigned short) pMemo->iBorderColor[1] ;
	bordercolor.blue  = (unsigned short) pMemo->iBorderColor[2] ;
	XAllocColor ( XtDisplay(wTop), colormap, &bordercolor ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNborderWidth, pMemo->iBorderWidth ) ; n ++ ;
	XtSetArg ( xArg[n], XtNborderColor, bordercolor.pixel ) ; n ++ ;
	pMemo->wMemo = XtAppCreateShell ( "xmemoslip", "XMemoSlip", overrideShellWidgetClass,
	                                  XtDisplay(wTop), xArg, n ) ;
	XtConfigureWidget ( pMemo->wMemo, pMemo->x, pMemo->y, pMemo->w, pMemo->h, pMemo->iBorderWidth ) ;

	/* Form */

	n = 0 ;
	XtSetArg ( xArg[n], XtNborderWidth    , 0 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNdefaultDistance, 0 ) ; n ++ ;
	pMemo->wForm = XtCreateManagedWidget ( "MemoForm", formWidgetClass, pMemo->wMemo, xArg, n ) ;

	/* Text */

	colormap = XDefaultColormap ( XtDisplay(wTop), 0 ) ;
	backcolor.red   = (unsigned short) pMemo->iBackColor[0] ;
	backcolor.green = (unsigned short) pMemo->iBackColor[1] ;
	backcolor.blue  = (unsigned short) pMemo->iBackColor[2] ;
	XAllocColor ( XtDisplay(wTop), colormap, &backcolor ) ;

	colormap = XDefaultColormap ( XtDisplay(wTop), 0 ) ;
	fontcolor.red   = (unsigned short) pMemo->iFontColor[0] ;
	fontcolor.green = (unsigned short) pMemo->iFontColor[1] ;
	fontcolor.blue  = (unsigned short) pMemo->iFontColor[2] ;
	XAllocColor ( XtDisplay(wTop), colormap, &fontcolor ) ;

	if ( pMemo->pFontName != NULL ) {
		pMemo->pFontStruct = XLoadQueryFont ( XtDisplay(wTop), pMemo->pFontName ) ;
		if ( pMemo->pFontStruct == NULL  &&  bMemoTraceLog ) {
			xmemo_TraceLog ( "xmemo_CreateMemoWidget", "XLoadQueryFont == NULL" ) ;
		}
	} else {
		pMemo->pFontStruct = NULL ;
	}

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth , pMemo->w - pMemo->iBorderWidth * 2 - 6 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNheight, pMemo->h - pMemo->iBorderWidth * 2 - 6 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNborderWidth, 0 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop   , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbottom, XtChainBottom ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft  , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright , XtChainRight  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNdisplayCaret, False   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbackground, backcolor.pixel ) ; n ++ ;
	XtSetArg ( xArg[n], XtNforeground, fontcolor.pixel ) ; n ++ ;
	if ( pMemo->pFontStruct != NULL ) {
		XtSetArg ( xArg[n], XtNfont, pMemo->pFontStruct ) ; n ++ ;
	}
	if ( pMemo->pText != NULL ) {
		XtSetArg ( xArg[n], XtNstring, pMemo->pText ) ; n ++ ;
	} else {
		XtSetArg ( xArg[n], XtNstring, "" ) ; n ++ ;
	}
	pMemo->wText = XtCreateManagedWidget ( "MemoText", asciiTextWidgetClass, pMemo->wForm, xArg, n ) ;

	XtAddEventHandler ( pMemo->wText, ButtonPressMask  , False, (XtEventHandler)MemoEvent_ButtonPress  , NULL ) ;
	XtAddEventHandler ( pMemo->wText, ButtonReleaseMask, False, (XtEventHandler)MemoEvent_ButtonRelease, NULL ) ;
	XtAddEventHandler ( pMemo->wText, Button1MotionMask, False, (XtEventHandler)MemoEvent_Drag         , NULL ) ;

	/* Border grip ( Bottom ) */

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth , pMemo->w - pMemo->iBorderWidth * 2 - 6 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNheight, 3 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNborderWidth, 0 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop   , XtChainBottom ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbottom, XtChainBottom ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft  , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright , XtChainRight  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert, pMemo->wText ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbackground, backcolor.pixel ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcursor, xBorderCursorBottom ) ; n ++ ;
	pMemo->wBorderBottom = XtCreateManagedWidget ( "Border", simpleWidgetClass, pMemo->wForm, xArg, n ) ;

	XtAddEventHandler ( pMemo->wBorderBottom, ButtonPressMask, False, (XtEventHandler)BorderEvent_ButtonPress, NULL ) ;
	XtAddEventHandler ( pMemo->wBorderBottom, Button1MotionMask, False, (XtEventHandler)BorderEvent_Drag, NULL ) ;

	/* Border grip ( Right ) */

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth , 3 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNheight, pMemo->h - pMemo->iBorderWidth * 2 - 6 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNborderWidth, 0 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop   , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbottom, XtChainBottom ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft  , XtChainRight  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright , XtChainRight  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, pMemo->wText ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbackground, backcolor.pixel ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcursor, xBorderCursorRight ) ; n ++ ;
	pMemo->wBorderRight = XtCreateManagedWidget ( "Border", simpleWidgetClass, pMemo->wForm, xArg, n ) ;

	XtAddEventHandler ( pMemo->wBorderRight, ButtonPressMask, False, (XtEventHandler)BorderEvent_ButtonPress, NULL ) ;
	XtAddEventHandler ( pMemo->wBorderRight, Button1MotionMask, False, (XtEventHandler)BorderEvent_Drag, NULL ) ;

	/* Border grip ( Bottom right ) */

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth , 3 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNheight, 3 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNborderWidth, 0 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop   , XtChainBottom ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbottom, XtChainBottom ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft  , XtChainRight  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright , XtChainRight  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, pMemo->wBorderBottom ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert , pMemo->wBorderRight  ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbackground, backcolor.pixel ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcursor, xBorderCursorCorner ) ; n ++ ;
	pMemo->wBorderCorner = XtCreateManagedWidget ( "Border", simpleWidgetClass, pMemo->wForm, xArg, n ) ;

	XtAddEventHandler ( pMemo->wBorderCorner, ButtonPressMask, False, (XtEventHandler)BorderEvent_ButtonPress, NULL ) ;
	XtAddEventHandler ( pMemo->wBorderCorner, Button1MotionMask, False, (XtEventHandler)BorderEvent_Drag, NULL ) ;

	/* Realize */

	XtRealizeWidget ( pMemo->wMemo ) ;

	return RC_SUCCESS ;

}

/******************************************************************************/
/*  xmemo_DestroyMemoWidget                                                   */
/******************************************************************************/

int xmemo_DestroyMemoWidget ( zMemoType *pMemo )
{

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_DestroyMemoWidget", NULL ) ;

	XtDestroyWidget ( pMemo->wMemo ) ;

	return RC_SUCCESS ;

}
